import os
import secrets
import unittest
import augpathlib as aug
from sparcur import exceptions as exc
from sparcur.utils import GetTimeNow
from sparcur.paths import PennsieveCache, LocalPath, Path
from sparcur.backends import PennsieveRemote
from .common import test_organization, test_dataset, _pid
from .common import skipif_ci, skipif_no_net
import pytest


class _TestOperation:

    _cache_class = PennsieveCache
    _remote_class = PennsieveRemote

    @classmethod
    def tearDownClass(cls):
        base = aug.AugmentedPath(__file__).parent / f'test-operation-{_pid}'
        if base.exists():
            base.popd()  # in case we were inside it pop back out first
            base.rmtree()

    def setUp(self):
        class Cache(self._cache_class):
            pass

        Cache._bind_flavours()

        base = aug.AugmentedPath(__file__).parent / f'test-operation-{_pid}'

        if base.exists():
            base.popd()  # in case we were inside it pop back out first
            base.rmtree()

        base.mkdir()
        base.pushd()

        self.Remote = self._remote_class._new(LocalPath, Cache)
        self.Remote.init(test_organization)
        self.anchor = self.Remote.dropAnchor(base)

        self.root = self.anchor.remote
        self.project_path = self.anchor.local
        list(self.root.children)  # populate datasets
        self.test_base = [
            p for p in self.project_path.children
            if p.cache.id == test_dataset][0]
        list(self.test_base.rchildren)  # populate test dataset

        asdf = self.root / 'lol' / 'lol' / 'lol (1)'

        class Fun(os.PathLike):
            name = 'hohohohohoho'

            def __fspath__(self):
                return ''

            @property
            def size(self):
                return aug.FileSize(len(b''.join(self.data)))

            @property
            def data(self):
                for i in range(100):
                    yield b'0' * 1000

        #wat = asdf.bfobject.upload(Fun(), use_agent=False)
        #breakpoint()

@skipif_ci
@skipif_no_net
class TestDelete(_TestOperation, unittest.TestCase):

    def test_0(self):
        assert True

    def test_1_case(self):
        # this is an old scenario that happens because of how the old system worked
        # local working directory | x
        # local cache directory | o
        # remote | o
        pass


def make_rand(n, width=80):
    lw = width + 1
    hex_lw = lw // 2
    n_lines = n // lw
    n_accounted = n_lines * lw
    h_lines = n // (hex_lw * 2)
    h_accounted = h_lines * (hex_lw * 2)
    ldiff = n_lines - h_lines
    adiff = n_accounted - h_accounted
    accounted = n_accounted
    missing = n - accounted
    hex_missing = (missing + 1) // 2
    diff = hex_missing * 2 - missing
    hexstart = width % 2  # almost there fails on 71
    # also fails len(make_rand(102, 101)) - 102

    print(adiff, ldiff, missing, diff)
    string = '\n'.join([secrets.token_hex(hex_lw)[hexstart:]
                        for i in range(n_lines)]
                       + [secrets.token_hex(hex_missing)[diff:-1] + '\n'])
    return string.encode()


@skipif_ci
@skipif_no_net
@pytest.mark.skip('VERY SLOW')
class TestFilenames(_TestOperation, unittest.TestCase):

    _evil_names = (
        # '......................',  # this breaks the agent with infinite timeout
        '!@#$%^&*()[]{}`~;:,',
        '(╯°□°）╯︵ ┻━┻)',
        'הָיְתָהtestالصفحات التّحول',
        'הָיְתָהtestالصفحا تالتّحول',  # check bucket names
        '᚛ᚄᚓᚐᚋᚒᚄ ᚑᚄᚂᚑᚏᚅ᚜',
        'Z̮̞̠͙͔ͅḀ̗̞͈̻̗Ḷ͙͎̯̹̞͓G̻O̭̗̮',
        # '𝕿𝖍𝖊 𝖖𝖚𝖎𝖈𝖐 𝖇𝖗𝖔𝖜𝖓 𝖋𝖔𝖝 𝖏𝖚𝖒𝖕𝖘 𝖔𝖛𝖊𝖗 𝖙𝖍𝖊 𝖑𝖆𝖟𝖞 𝖉𝖔𝖌',  # this breaks the agent with ERRORED
        'evil file space',
        'evil_file underscore',
        'evil-file dash',
        'evil%20file percent 20',
        'hello%20world%20%60~%21%40%23%24%25%5E%26%2A%28%29%5B%5D%7B%7D%27',
        # the problem is that we don't know whether we can actually
        # decode a file name, and wthe database stores the encoded filename
        'hello%20world',
        'hello%20world~',
    )

    @property
    def _more_evil_names(self):
        # class scope strikes back! LOL PYTHON
        return [
            name
          for char in
            ('\x07',
             #'/',  # have to do this in a different way on unix
             '\\',
             '|',
             '!',
             '@',
             '#',
             '$',
             '%',
             '^',
             '&',
             '*',
             '(',
             ')',
             '[',
             ']',
             '{',
             '}',
             "'",
             '`',
             '~',
             ';',
             ':',
             ',',
             '"',
             '?',
             '<',
             '>',
             )
          for name in
            (f'prefix{char}',
             f'prefix{char}suffix',
             f'{char}suffix',)]

    @staticmethod
    def _op(test_folder, fsize, name):
        test_file_a = test_folder / (name + '.ext')
        test_file_b = test_folder / (name + '.txe')
        test_folder_i = test_folder / name

        for _f in (test_file_a, test_file_b):
            if _f.exists() or _f.is_broken_symlink():
                msg = (f'bad test environment: file/link already exists: {_f}')
                raise FileExistsError(msg)

        # FIXME maybe don't straight fail here, but instead
        # don't upload and just compare the existing name?
        # the fact that we get an error is a sign that the
        # name matches actually ... so not getting an error
        # in subsequent runs is bad ... for test_base at least

        test_file_a.data = iter((make_rand(fsize),))
        test_file_b.data = iter((make_rand(fsize),))

        try:
            remote_a = test_file_a.upload()
            name_a = remote_a.bfobject.name
        except Exception as e:
            name_a = e

        try:
            remote_b = test_file_b.upload()
            name_b = remote_b.bfobject.name
        except Exception as e:
            name_b = e

        try:
            remote_i = test_folder_i.mkdir_remote()
            name_i = remote_i.bfobject.name
        except Exception as e:
            name_i = e

        return name_a, name_b, name_i

    def test_filenames_more_evil(self):
        return self.test_filenames_evil(self._more_evil_names)

    def test_filenames_evil(self, names=_evil_names):
        # XXX warning slow!
        now = GetTimeNow()
        local = self.project_path / f'test-dataset-{now.START_TIMESTAMP_LOCAL_FRIENDLY}'
        remote = local.mkdir_remote()

        try:
            # FIXME consider going back to self.test_base instead of local here
            test_folder = local / 'pandora'
            test_folder.mkdir_remote()
            test_folder.__class__.upload = Path.upload
            results = []
            fsize = 1024  # needed for uniqueish hashes colloisions will still happen
            # FIXME this pretty clearly reveals a need for
            # batching to multiplex the fetch ... SIGH
            for name in names:
                name_a, name_b, name_i = self._op(test_folder, fsize, name)
                results.append((name_a, name_b, name_i))
        finally:
            remote.rmdir()
            remote.cache.refresh()


@skipif_ci
@skipif_no_net
class TestUpdate(_TestOperation, unittest.TestCase):

    @pytest.mark.skip('the question has been answered')
    def test_process_filesize_limit(self):
        # so the 1 mb size file works, eventually, something else is wrong
        test_folder = self.test_base / 'hrm'
        test_folder.mkdir_remote()
        test_folder.__class__.upload = Path.upload
        for i in range(1024 ** 2, 5 * 1024 ** 2, 1024 ** 2):
            test_file = test_folder / f'size-{i}'
            if test_file.is_broken_symlink():
                test_file.remote.bfobject.package.delete()
                test_file.unlink()
                test_file = test_folder / f'size-{i}'  # remove stale cache

            test_file.data = iter((make_rand(i),))
            remote = test_file.upload()

    def test_upload_noreplace(self):
        for i in range(2):
            test_file = self.test_base / 'dataset_description.csv'
            test_file.data = iter((make_rand(100),))
            # FIXME temp sandboxing for upload until naming gets sorted
            test_file.__class__.upload = Path.upload
            # create some noise
            remote = test_file.upload(replace=False)
            print(remote.bfobject.package.name)

    def test_upload_noreplace_fail(self):
        # some persistent state from other tests is causing this to fail
        test_file = self.test_base / 'dataset_description.csv'
        test_file.data = iter((make_rand(100),))
        # FIXME temp sandboxing for upload until naming gets sorted
        test_file.__class__.upload = Path.upload
        test_file.upload(replace=False)
        try:
            test_file.upload(replace=False)
            assert False, 'should have failed'
        except exc.FileHasNotChangedError:
            pass

    def test_upload_replace(self):
        test_file = self.test_base / 'dataset_description.csv'
        test_file.data = iter((make_rand(100),))
        # FIXME temp sandboxing for upload until naming gets sorted
        test_file.__class__.upload = Path.upload
        test_file.upload()


@skipif_ci
@skipif_no_net
class TestClone(_TestOperation, unittest.TestCase):
    # TODO test a variety of clone scenarios
    # and consider whether testing for and
    # existing root should be done in dropAnchor

    def setUp(self):
        super().setUp()
        self.alt_project_path = self.project_path.parent / 'alt' / self.project_path.name
        if self.alt_project_path.parent.exists():
            self.alt_project_path.parent.rmtree()

        self.alt_project_path.mkdir(parents=True)

    def _do_target(self, target, expect_error_type=None):
        class Cache(self._cache_class):
            pass

        Cache._bind_flavours()

        BFR = self._remote_class._new(LocalPath, Cache)
        BFR.init(test_organization)

        if expect_error_type:
            try:
                anchor = BFR.dropAnchor(target.parent)
                raise AssertionError(f'should have failed with a {expect_error_type}')
            except expect_error_type as e:
                pass

        else:
            anchor = BFR.dropAnchor(target.parent)

    def test_1_in_project(self):
        target = self.project_path / 'some-new-folder'
        target.mkdir()
        self._do_target(target)  # FIXME succeeds for now, but probably should not?

    def test_2_project_top_level(self):
        target = self.project_path
        self._do_target(target, exc.DirectoryNotEmptyError)

    def test_3_existing_empty(self):
        target = self.alt_project_path
        self._do_target(target)

    def test_4_existing_has_folder(self):
        target = self.alt_project_path
        child = target / 'a-folder'
        child.mkdir(parents=True)
        self._do_target(target, exc.DirectoryNotEmptyError)

    def test_5_existing_has_file(self):
        target = self.alt_project_path
        child = target / 'a-file'
        child.touch()
        self._do_target(target, exc.DirectoryNotEmptyError)

    def test_6_existing_has_local_data_dir(self):
        target = self.alt_project_path
        child = target / self.anchor._local_data_dir
        child.mkdir()
        self._do_target(target, exc.DirectoryNotEmptyError)


@skipif_ci
@skipif_no_net
class TestMkdirRemote(_TestOperation, unittest.TestCase):

    def test_mkdir_remote_parents_false(self):
        now = GetTimeNow()
        local = self.project_path / f'test-dataset-{now.START_TIMESTAMP_LOCAL_FRIENDLY}' / 'some-folder'
        try:
            remote = local.mkdir_remote()
            raise AssertionError('Should have failed since parents=False')
        except FileNotFoundError:
            pass

    def test_0_mkdir_remote_will_be_dataset(self):
        now = GetTimeNow()
        local = self.project_path / f'test-dataset-{now.START_TIMESTAMP_LOCAL_FRIENDLY}'
        remote = local.mkdir_remote()
        remote.rmdir()
        remote.cache.refresh()  # reminder that remotes are a snapshot in time, NOT dynamic
        assert not local.exists(), f'should have been deleted {remote}'

    def test_1_mkdir_remote_will_be_collection(self):
        now = GetTimeNow()
        local = self.project_path / f'test-dataset-{now.START_TIMESTAMP_LOCAL_FRIENDLY}' / 'some-folder'
        remote = local.mkdir_remote(parents=True)
        parent = remote.parent
        try:
            parent.rmdir()  # should fail here
            try:
                remote.rmdir()  # insurance
            except BaseException as e:
                log.exception(e)
            finally:
                raise AssertionError('remote parent should NOT have rmdired {parent}')
        except exc.PathNotEmptyError:
            pass

        try:
            remote.rmdir()
            remote.cache.refresh()
            assert not local.exists(), f'should have been deleted {remote}'
        finally:
            lparent = parent.local
            parent.cache.refresh()  # we just removed the child so the parent is stale
            parent.rmdir()
            parent.cache.refresh()
            assert not lparent.exists(), f'should have been deleted {parent}'



class TestRenameFolder:
    # TODO
    pass


class TestMoveFolder:
    # TODO
    pass


class TestRemote(_TestOperation, unittest.TestCase):
    def test_remote_path_does_not_exist(self):
        new_thing = self.root / 'does not exist'

    @pytest.mark.skip('Not ready.')
    def test_cache_path_does_not_exist(self):
        """ This should not produce an error.
            Path objects should be able to be instantiated without
            po.exists() -> True at a point in time prior to instantiation.
        """
        new_thing = self.anchor / 'does not exist'

    def __test_cache_path_fake_id(self):
        # FIXME the right way to do this is
        np = self.project_path / 'new-path'
        np.mkdir()
        npm = np.meta
        
        # bad way
        class FakeBase(aug.RemotePath):
            def __init__(self, id, name, cache=None):
                super().__init__(id, cache)
                self.name = name
                now = GetTimeNow()
                self.created = now._start_time
                self.updated = now._start_time
                self.checksum = 'lolnone'
                self.chunksize = 4096
                self.file_id = 'asdfasdfasdf'

            @property
            def meta(self):
                return PathMeta(size=self.size,
                                created=self.created,
                                updated=self.updated,
                                checksum=self.checksum,
                                chunksize=self.chunksize,
                                id=self.id,
                                file_id=self.file_id)


            @property
            def parent(self):
                return None

            def _parts_relative_to(self, remote, cache_parent=None):
                return [self.name]


        # This takes way too many steps :/
        Fake = FakeBase._new(LocalPath, aug.CachePath)

        fake = Fake('lol', 'double lol')

        self.anchor / fake





import idlib
import ontquery as oq
import augpathlib as aug
from github import Github
from pyontutils import obo_io as oio
from pyontutils import sheets
from pyontutils.core import OntResIri, OntId
from pyontutils.config import auth
from pyontutils.namespaces import ilxtr

# for reference see
# https://github.com/obophenotype/uberon/blob/master/uberon-idranges.owl
# Datatype: idrange:22
# Annotations: allocatedto: "InterLex",
# EquivalentTo: xsd:integer[> 8310000, <= 8399999]
# hard coding due to lack of a manchester syntax parser
uberon_id_range = 8310000, 8400000


def github_url_to_group_repo(url_string):
    # FIXME github_url_to_python_object is what we really want
    # but that requires a bit more setup via orthauth
    from urllib.parse import urlparse
    url_obj = urlparse(url_string)
    netloc = url_obj.netloc
    if netloc in ('github.com',  'raw.githubusercontent.com'):
        group_repo = '/'.join(url_obj.path.split('/', 3)[1:3])
        return group_repo
    else:
        raise ValueError(f"Don't know how to handle uris for {netloc}")


class Hrm(OntResIri):

    def submit_terms_to(self, upstream_iri, *identifiers, **kwargs):
        pass


class OntTerm(idlib.from_oq.OntTerm):

    def asOboTerm(self, *args, id='tgbugsTODO'):
        graph = self.query.services[0].graph
        ilx_id = OntId(self).u
        s = list(graph[:ilxtr.hasIlxId:ilx_id])[0]
        if s != ilx_id:
            self = self.__class__(s)
            self.fetch()

        try:
            term = oio.Term(id=id, name=self.label)
            if not self.definition:
                raise ValueError(f'{self} is misisng a definition!')

            defxrefs = list(graph[s:OntId('ilx.anno.hasDefinitionSource:').u:])
            term.add(oio.TVPair(tag='def', text=self.definition, xrefs=defxrefs))

            for synonym in (self.synonyms
                            if isinstance(self.synonyms, tuple) else  # FIXME ontquery issue
                            (self.synonyms,)):
                term.add(oio.TVPair(tag='synonym', text=synonym, typedef='EXACT'))

            # FIXME SIGH brokenness of OntTerm.__call__ return types
            ps = ('ilx.anno.hasBroadSynonym:', 'oboInOwl:hasBroadSynonym')
            broads = [v for k, vs in self(*ps).items() if k in ps for v in vs]
            for bs in broads:
                term.add(oio.TVPair(tag='synonym', text=bs, typedef='BROAD'))

            ps = ('ilx.anno.hasRelatedSynonym:', 'oboInOwl:hasRelatedSynonym')
            relateds = [v for k, vs in self(*ps).items() if k in ps for v in vs]
            for rs in relateds:
                term.add(oio.TVPair(tag='synonym', text=rs, typedef='RELATED'))

            for iri in graph[s:ilxtr.hasExistingId]:
                term.add(oio.TVPair(tag='xref', name=OntId(iri).curie,))

        except (AttributeError, TypeError) as e:
            raise ValueError(self) from e

        return term


class Row(sheets.Row):

    _repo_objs = {}
    _pr_objs = {}

    @classmethod
    def github_api(cls):
        if not hasattr(cls, '_github_api'):
            cls._github_api = Github(auth.user_config.secrets(
                'github',
                'tgbugs-build',  # FIXME obvs
                'WARNING-development-all-scopes'))

        return cls._github_api

    @classmethod
    def pr_repo(cls, group_repo):
        if group_repo not in cls._repo_objs:
            gh_repo_obj = cls.github_api().get_repo(group_repo)
            cls._repo_objs[group_repo] = gh_repo_obj
        return cls._repo_objs[group_repo]

    @property
    def id(self):
        return OntTerm(self.interlex_id().value)

    @property
    def ready(self):
        return self.ready_to_go().value

    @property
    def submitted(self):
        test = 'https://github.com/obophenotype/uberon/pull/'
        return self.pull_request_link().value.startswith(test)
        #return self.status().value == 'pr submitted'

    @property
    def merged(self):
        prv = self.pull_request_link()
        url = prv.value
        group_repo = github_url_to_group_repo(url)
        repo = self.pr_repo(group_repo)
        _, pr_number_string = url.rsplit('/', 1)
        pr_number = int(pr_number_string)
        # lol github using the same index for prs an issues useful to
        # simplify referencing things, but means that you don't know
        # the type without checking it when you only have the number
        if pr_number not in self._pr_objs:
            self._pr_objs[pr_number] = repo.get_pull(pr_number)

        pr = self._pr_objs[pr_number]
        return pr.merged

    @property
    def to_submit(self):
        return self.ready and not self.merged  # submitted terms may need to be updated


# monkey patch
sheets.Row = Row


class UpstreamTermRequests(sheets.Sheet):
    name = 'sparc-upstream-terms'
    sheet_name = 'upstream'
    index_columns = 'interlex_id',

    def _row_objects(self):
        ros = [self.row_object(i) for i, v  in enumerate(self.values)][1:]
        return ros

    def ready(self):
        return [r for r in self._row_objects() if r.ready]
    
    def submitted(self):  # FIXME should be on row
        return [r for r in self._row_objects() if r.submitted]

    def merged(self):  # FIXME should be on row
        return [r for r in self._row_objects() if r.merged]

    def to_submit(self):
        # TODO actually check uberon
        return [r for r in self._row_objects() if r.to_submit]

    def submit_to_obofile(self, of, prefix, id_range):
        terms = [ro.id for ro in self.to_submit()]
        [t.fetch() for t in terms]
        #obo_terms = [t.asOboTerm(id=f'tgbugsTODO{i}') for i, t in enumerate(terms)]
        id_min, id_max = id_range
        over_under = [
            int(b.id_.value.suffix)
            for b in of.Terms.values()
            if not isinstance(b, list) and ':' in b.id_.value
            and id_min < int(OntId(b.id_.value).suffix) < id_max]
        if over_under:
            id_start = max(over_under) + 1
        else:
            id_start = id_min

        # TODO padding and prefix detect etc.
        obo_terms = [t.asOboTerm(id=f'{prefix}:{id_start + i}') for i, t in enumerate(terms)]
        of.add(*obo_terms)


def main():
    #ori = OntResIri('https://alt.olympiangods.org/sparc/ontologies/community-terms.ttl')
    ori = OntResIri('http://localhost:8515/sparc/ontologies/community-terms.ttl')
    rdfl = oq.plugin.get('rdflib')(ori.graph, OntId)
    OntTerm.query_init(rdfl)
    utr = UpstreamTermRequests()
    _todo(utr)
    return

    # test output
    tof = oio.OboFile()
    utr.submit_to_obofile(tof, 'FIXME', (0, float('inf')))
    tof.write(path=aug.RepoPath('/tmp/uberon-new-terms.obo'), overwrite=True)

    ori_uberon = OntResIri('http://purl.obolibrary.org/obo/uberon.owl')
    ori_uberon_meta = ori_uberon.metadata()  # doap:GitRepository
    breakpoint()
    return


def _todo(utr):
    # real output
    glb = auth.get_path('git-local-base')
    uberon_edit = aug.RepoPath(glb) / 'NOFORK/uberon/src/ontology/uberon-edit.obo'
    of = oio.OboFile(path=uberon_edit, strict=False)
    utr.submit_to_obofile(of, 'UBERON', uberon_id_range)
    of.write(overwrite=True,
             version=oio.OBO_VER_ROBOT)

    # TODO 
    # source-iri + identifiers -> target-iri
    # need a target-iri -> edit type + edit id
    # e.g for uberon if there is a branch where we have stuck our working pull requests
    # that goes in the edit id.
    # https://alt.olympiangods.org/sparc/ontologies/community-terms.ttl + ids ->
    # http://purl.obolibrary.org/obo/uberon.owl -> 
    # https://github.com/obophenotype/uberon/tree/sparc-term-request-flow/blob/uberon_edit.obo


if __name__ == '__main__':
    main()



import io
import csv
import copy
from types import GeneratorType
from itertools import chain
from collections import Counter, defaultdict
#import openpyxl  # import time hog
import augpathlib as aug
from xlsx2csv import Xlsx2csv, SheetNotFoundException, InvalidXlsxFileException
from terminaltables import AsciiTable
from pyontutils.utils import python_identifier
from . import schemas as sc
from . import raw_json as rj
from . import exceptions as exc
from . import normalization as nml
from .core import log, logd, HasErrors
from .paths import Path, PennsieveCache
from .utils import is_list_or_tuple


# FIXME review this :/ it is not a good implementation at all
def tos(hrm):
    if isinstance(hrm, GeneratorType):
        t = tuple(hrm)
        if not t:
            return
        elif len(t) == 1:
            return t[0]
        else:
            return t
    else:
        return hrm


def to_string_and_then_python_identifier(thing):
    """ handl cases where a header value isn't a string """
    return python_identifier(str(thing))


hasSchema = sc.HasSchema()
@hasSchema.mark
class Header:
    """ generic header normalization for python """
    # FIXME this is a really a pipeline stage
    def __init__(self, first_row_or_column, normalize=True, alt=False, normalize_first_cell=False):
        self.pipeline_start = first_row_or_column
        self._normalize = normalize
        self._alt = alt
        self._normalize_first_cell = normalize or normalize_first_cell

    @property
    def normalized(self):
        if not hasattr(self, '_cache_normalized'):
            orig_header = self.pipeline_start
            header = []
            prefix = 'TEMPA' if self._alt else 'TEMPH' # prevent collision
            for i, c in enumerate(orig_header):
                if c:
                    c = to_string_and_then_python_identifier(c)
                    c = nml.NormHeader(c)

                if not c:
                    c = f'{prefix}_{i}'

                if c in header:
                    c = c + f'_{prefix}_{i}'

                header.append(c)

            self._cache_normalized = header

        return self._cache_normalized

    @hasSchema(sc.HeaderSchema)
    def data(self):
        return self.normalized

    @property
    def _lut(self):
        return {n:h for n, h in zip(self.data, self.pipeline_start)}

    def lut(self, **renames):
        if self._normalize:
            return self._lut_normalize(renames)
        else:
            return self._lut_original(renames)

    def _lut_normalize(self, renames):
        """ do any renames beyond the specific normalization """
        return {renames[n] if n in renames else n:h
                for n, h in zip(self.data, self.pipeline_start)}

    def _lut_original(self, renames):
        original = self.pipeline_start
        normalized = self.normalized
        if len(set(original)) != len(original):
            dupes = [v for v, c in Counter(original).most_common() if c > 1]
            msg = f'Original header is not unique.\nDuplicate entries: {dupes}'
            if None in dupes:
                msg += '\nThere may be multiple records without a primary key.'
            raise exc.MalformedHeaderError(msg)

        if self._normalize_first_cell:
            norm_first = (*normalized[:1], *original[1:])
            return {renames[n] if n in renames else n:h
                    for n, h in zip(norm_first, original)}
        else:
            return {renames[n] if n in renames else n:h
                    for n, h in zip(original, original)}


class DatasetMetadata(HasErrors):

    def __init__(self, path, template_schema_version=None):
        self._path = path
        self.cache = path.cache
        self.name = path.name
        # XXX transitive updated is not entirely necessary anymore since the
        # dataset updated date is bumped on any transitive modification
        self.updated_cache_transitive = path.updated_cache_transitive
        self.template_schema_version = template_schema_version

    @property
    def data(self):
        if self.cache is not None:
            cmeta = self.cache.meta
            return dict(
                id=self.cache.id,
                meta=dict(
                    folder_name=self.name,
                    uri_human=self.cache.uri_human,
                    uri_api=self.cache.uri_api,
                    timestamp_created=cmeta.created,
                    timestamp_updated=cmeta.updated,
                    # FIXME vs the dataset updated time now touched for transitive updates
                    timestamp_updated_contents=self.updated_cache_transitive(),
                    sparse=self.cache.is_sparse(),
                        ))
        else:
            raise NotImplementedError('use CacheL instead')
            path = self._path
            meta = path.meta
            #uri = quote(path.resolve().as_uri(), safe=':/')
            uri = path.resolve().as_uri()
            return dict(id=meta.id,
                        meta=dict(folder_name=path.name,
                                  uri_human=uri,
                                  uri_api=uri,
                                  timestamp_created=None,  # cannot get birthdate on all file systems
                                  timestamp_updated=meta.updated,
                        ))


class DatasetStructure:
    sections = None  # NOTE assigned in at end of file
    rglobs = 'manifest',
    default_glob = 'glob'
    max_childs = None  # 40 was used when we didn't have sparse pulls
    rate = 8  # set by Integrator from cli
    _refresh_on_missing = True
    _classes = {}

    def _newimpl(cls, path, *args, **kwargs):
        apc = path.__class__._AugmentedPath__abstractpath
        if apc not in cls._classes:
            # keep classes 1:1 with the types of paths so that equality works correctly
            # mro issues with specialized paths instead of the base
            newcls = type(cls.__name__ + apc.__name__, (cls, apc),
                          #cls.__dict__,
                          dict(__new__=apc.__new__,
                               __repr__=apc.__repr__,
                               )
                          )
            newcls._bind_flavours(pos_helpers=(HasErrors,), win_helpers=(HasErrors,))
            cls._classes[apc] = newcls

        nc = cls._classes[apc]
        return nc(path)

    def __new__(cls, path, *args, **kwargs):
        return cls._newimpl(cls, path)

    _renew = __new__

    def __new__(cls, path, *args, **kwargs):
        DatasetStructure._setup()
        DatasetStructure.__new__ = DatasetStructure._renew

        return cls._newimpl(cls, path)

    @staticmethod
    def _setup():
        import requests
        DatasetStructure._requests = requests

    @property
    def children(self):
        # for whatever reason the way pathlib constructs children
        # skips calling init, iirc because they use object.__new__
        # or something

        for child in super().children:
            child.__init__()
            yield child

    @classmethod
    def _bind_sections(cls):
        for section, _sec_class in cls.sections.items():
            @property
            def sec(self, sec_class=_sec_class, name='_cache_' + _sec_class.__name__):
                if not hasattr(self, name):
                    mp = [p for p in self.meta_paths if isinstance(p, sec_class)]
                    if len(mp) == 1:
                        setattr(self, name, mp[0])
                    elif not mp:
                        setattr(self, name, None)
                    else:
                        setattr(self, name, mp)

                return getattr(self, name)

            setattr(cls, section, sec)

    def files_last_updated(self):
        # FIXME TODO can we get this more efficiently during pull?
        all_updated = [c.cache.meta.updated for c in
                       chain((self,), self.rchildren) if c.is_file()]
        if all_updated:
            return max(all_updated)

    @property
    def counts(self):
        if not hasattr(self, '_counts'):
            size = 0
            dirs = 0
            files = 0
            need_meta = []
            if not self.is_dir():
                gen = self,

            else:
                gen = self.rchildren

            for c in gen:
                if c.is_dir():
                    dirs += 1
                    maybe_size = 0  # the remote size of a folder is zero for bf
                else:
                    files += 1  # testing for broken symlinks is hard
                    try:
                        if c.is_broken_symlink():
                            maybe_size = (aug.PathMeta
                                          .from_symlink(c)
                                          .size)
                        else:
                            _meta = aug.PathMeta.from_xattrs(c.xattrs(),
                                                             prefix='bf')
                            if _meta.size is not None:
                                maybe_size = _meta.size
                            elif c.is_file():  # running an export on a local dataset
                                maybe_size = c.size

                        #maybe_size = c.cache.meta.size
                    except AttributeError as e:
                        log.error(f'no cache or no meta for {c}\n{e}')
                        continue

                    if maybe_size is None:
                        need_meta.append(c)
                    else:
                        size += maybe_size

            if need_meta and self._refresh_on_missing:
                msg = "We don't do this anymore. Fetch everything first."
                raise exc.NetworkSandboxError(msg)

            self._counts = dict(size=aug.FileSize(size), dirs=dirs, files=files)

        return self._counts

    @property
    def total_size(self):
        """ total remote size, trigger retrieval of all remote metadata """
        return self.counts['size']

    @property
    def total_dirs(self):
        return self.counts['dirs']

    @property
    def total_files(self):
        return self.counts['files']

    @property
    def total_paths(self):
        return self.counts['dirs'] + self.counts['files']

    @property
    def dataset_root(self):
        # FIXME this will find the first dataset description file at any depth!
        # this is incorrect behavior!
        """ The folder where the dataset_description file is located.
            Sometimes there is an intervening folder. """
        if self.cache.is_dataset:
            def check_fordd(paths, level=0, stop=3):
                if not paths:  # apparently the empty case recurses forever
                    return

                if (self.max_childs is not None and
                    self.max_childs > 0 and
                    len(paths) > self.max_childs):
                    log.warning(f'Not globing in a folder with > {self.max_childs} children! '
                                f'{self.as_posix()!r}')
                    return
                dd_paths_all = []
                children = []
                for path in paths:
                    dd_paths = list(path.glob('[Dd]ataset_description*.*'))
                    if dd_paths:
                        dd_paths_all.extend(dd_paths)
                    elif not dd_paths_all:
                        children.extend([p for p in path.children if p.is_dir()])

                if dd_paths_all:
                    return dd_paths_all
                else:
                    return check_fordd(children, level + 1)

            dd_paths = check_fordd((self,))

            if not dd_paths:
                #log.warning(f'No bids root for {self.name} {self.id}')  # logging in a property -> logspam
                return

            elif len(dd_paths) > 1:
                #log.warning(f'More than one submission for {self.name} {self.id} {dd_paths}')
                pass

            return dd_paths[0].parent  # FIXME choose shorter version? if there are multiple?

        elif self.parent:  # organization has no parent
            return self.parent.dataset_root

    def _abstracted_paths(self, name_prefix, glob_type=None, fetch=True, sandbox=False):
        """ A bottom up search for the closest file in the parent directory.
            For datasets, if the bids root and path do not match, use the bids root.
            In the future this needs to be normalized because the extra code required
            for dealing with the intervening node is quite annoying to maintain.
        """
        if glob_type is None:
            glob_type = self.default_glob

        path = self
        cache = self.cache
        if cache and cache == cache.anchor:  # organization
            for child in self.children:
                yield from child._abstracted_paths(name_prefix,
                                                   glob_type,
                                                   fetch,
                                                   sandbox,)
            return
        elif (self.cache and
              self.cache.is_dataset and
              self.dataset_root is not None and
              self.dataset_root != self):
            path = self.dataset_root

        first = name_prefix[0]
        cased_np = '[' + first.upper() + first + ']' + name_prefix[1:]  # FIXME warn and normalize
        glob = getattr(path, glob_type)
        gen = glob(cased_np + '*.*')

        try:
            path_class = self.sections[name_prefix]
            path = next(gen)
            for path in chain((path,), gen):
                path = path_class(path)
                if path.is_broken_symlink():
                    if not fetch:
                        yield path
                        continue
                    elif sandbox:
                        msg = ('A required metadata path was not fetched '
                               f'in advance. {path}')
                        raise exc.NetworkSandboxError(msg)

                    log.info(f'fetching unretrieved metadata path {path.as_posix()!r}'
                             '\nFIXME batch these using async in cli export ...')
                    try:
                        path.cache.fetch(size_limit_mb=path.cache.meta.size.mb + 1)
                    except self._requests.exceptions.ConnectionError as e:
                        raise exc.NetworkFailedForPathError(path) from e

                name = path.name
                if (path.is_file() and
                    (name.startswith('.~lock.') and name.endswith('#')) and
                    path.cache.meta.id is None):
                    # FIXME abstract to handle arbitrary path name regex patterns
                    log.warning(f'illegal untracked file detect {path}')
                    continue

                if path.suffix in path.stem:
                    msg = f'path has duplicate suffix {path.as_posix()!r}'
                    self.addError(msg,
                                  blame='submission',
                                  path=path)
                    logd.error(msg)

                if name[0].isupper():  # FIXME should probably be checking name.lower() != name
                    msg = f'path has bad casing {path.as_posix()!r}'
                    if self.addError(msg,
                                     blame='submission',
                                     path=path):
                        logd.error(msg)

                yield path

        except StopIteration:
            if (self.cache.parent.meta is not None and
                self.parent.cache != self.cache.anchor and
                self.parent != self):
                # FIXME 99% sure that this never triggers during normal use
                # if you hit this there is a good chance that the path you
                # fed to the DatasetStructure is not a dataset path which
                # will break things
                yield from getattr(self.parent, name_prefix + '_paths')

    @property
    def meta_paths(self):
        for section_name in self.sections:
            glob_type = 'rglob' if section_name in self.rglobs else None
            yield from self._abstracted_paths(section_name, glob_type=glob_type)

    @property
    def data(self):
        # XXX beware if you ever try to call checksum on this it will
        # fail because this is no longer a path in the way that the path
        # checksum exists because this data returns a dict, this could
        # be remedied by coming up with a convention for hashing arbitrary
        # python objects, but we aren't there yet

        out = self.counts
        # out['paths'] = self.inverted_index()
        for section_name in self.sections:
            #section_paths_name = section_name + '_paths'
            #paths = list(getattr(self, section_paths_name))
            glob_type = 'rglob' if section_name in self.rglobs else 'glob'
            paths = list(self._abstracted_paths(section_name, glob_type=glob_type))
            section_key = section_name + '_file'  # XXX _file is added here
            if paths:
                if len(paths) == 1 and section_name != 'manifest':  # FIXME manifest is a hack
                    path, = paths
                    out[section_key] = path
                else:
                    out[section_key] = paths

        out['dir_structure'] = list(self.data_dir_structure().values())

        self.embedErrors(out)
        return out

    xattr_prefix = PennsieveCache.xattr_prefix
    cache_meta = aug.EatCache.meta

    def data_dir_structure(self):
        #return {d.dataset_relative_path:d.cache_meta for d in self.rchildren_dirs}
        # TODO get all the leaves and then build an inverted index
        def jmc(d): return d.cache._jsonMetadata()
        def jml(d): return d._jsonMetadata()  # XXX TODO FIXME for local validation w/o cache

        def wrap_pathmeta_error(e):
            return [{'message': e,
                     'blame': 'remote',
                     'pipeline_stage': 'DatasetStructure.data_dir_structure',}]
        # FIXME this is just the directories so it is probably ok to
        # resolve the cache for now so we can get the nice json
        # metadata but ultimately we will need _cache_jsonMetadata or similar
        # that uses xattrs to avoid the overhead of constructing the cache class
        # XXX yep, _cache_jsonMetadata is needed especially for standalone validation
        # XXX note that this does NOT solve the issue where there are files without
        # cached metadata in a dataset that was derived from a remote source
        jsonMeta = jmc if self.cache is not None else jml
        instpath = self._cache_class._local_class  # XXXXXXXXXXX FIXME oof inheritance :/
        return {d.dataset_relative_path:{
            **jsonMeta(d),
            'path_meta': {k:wrap_pathmeta_error(v) if k == 'errors' else v
                          for k, v in self.__class__(d).cache_meta.items()
                          # errors can be strings in here! they come from
                          # backends.BlackfynnRemote.meta (for example)
                          if v is not None and (v != tuple()
                                                if k == 'errors'
                                                else True) }}
                for d in (instpath(d) for d in self.rchildren_dirs)}

    def inverted_index(self):
        # TODO there are two steps here
        # 1. identifying invalid folder hierarchies
        #    (e.g. materialized subjects and samples folders)
        # 2. _after_ we have the subjects identifiers we
        #    need to find all the matches remove all the nested
        #    non-matching and then recheck for invalid hierarchies
        # at the moment there seem to be lots of incorrect structures
        # where the primary has a bunch of stuff in between

        # {sourc,primary,derivative}/{subject_id,sample_id,pool_id,performance_id}/{don-t-really-care}...
        dds = self.data_dir_structure()
        #dd = defaultdict(lambda : defaultdict(list))
        dd = defaultdict(list)
        for d in dds:
            p = d.parts
            dd[p[-1]].append((len(p), d, p[::-1]))

        return dict(dd)


class ObjectPath:

    obj = None
    _classes = {}

    _newimpl = DatasetStructure._newimpl

    def __new__(cls, path):
        return cls._newimpl(cls, path)

    @property
    def object(self):
        return self.obj(self)

    def object_new(self, template_schema_version=None):
        """ PROPERTIES ARE BAD KIDS DON'T DO IT UNLESS
            YOU ARE REALLY SURE """  # sigh

        return self.obj(self, template_schema_version)


class Tabular(HasErrors):

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)

    _renew = __new__

    def __new__(cls, *args, **kwargs):
        # wouldn't it be great if you could bind tabular
        # to _CLASS=Tabular in new and then not have to
        # change the contents of the code and risk difficult
        # to debug copy/paste errors ? LOL PYTHON sigh
        Tabular._setup(*args, **kwargs)
        Tabular.__new__ = Tabular._renew
        return super().__new__(cls)

    @staticmethod
    def _setup(*args, **kwargs):
        import openpyxl
        Tabular._openpyxl = openpyxl

    def __init__(self, path, rotate=False):
        super().__init__()
        self.path = path
        self._rotate = rotate

    @property
    def file_extension(self):
        if self.path.suffixes:
            ext = self.path.suffixes[0]  # FIXME filenames with dots in them ...
            if ext != '.fake':
                return nml.NormFileSuffix(ext).strip('.')

    def tsv(self):
        return self.csv(delimiter='\t')

    def csv(self, delimiter=','):
        # FIXME if the number of delimiters in a row other than the header
        # is greater than the number of delimiters in the header then any
        # columns beyond the number in the header will be truncated and
        # I'm not sure that this loading routine detects that
        for encoding in ('utf-8', 'latin-1'):
            try:
                with open(self.path, 'rt', encoding=encoding) as f:
                    rows_orig = list(csv.reader(f, delimiter=delimiter))

                rows = [row for row in rows_orig if row and any(row)]
                diff = len(rows_orig) - len(rows)
                if diff:
                    # LOL THE FILE WITH > 1 million empty rows, 8mb of commas
                    # totally the maximum errors champion
                    message = f'There are {diff} empty rows in {self.path.as_posix()!r}'
                    if self.addError(message,
                                     blame='submission',
                                     path=self.path):
                        logd.error(message)

                if not rows:
                    ps = self.path.size
                    pcs = self.path.cache.meta.size
                    if ps != pcs:
                        msg = ('We have a likely case of a file that '
                               f'failed to fetch {ps} != {pcs}! {self.path}')
                        log.critical(msg)
                        self.addError(msg,
                                      blame='fetch',
                                      path=self.path)

                    return

                if len(rows[0]) == 1 and ('\t' if delimiter == ',' else ',') in rows[0][0]:
                    message = (f'Possible wrong file extension in {self.path.id} '
                               f'{self.path.project_relative_path}')
                    if self.addError(message,
                                     blame='submission',
                                     path=self.path):
                        logd.error(message)

                if encoding != 'utf-8':
                    message = f'encoding bad {encoding!r} {self.path.as_posix()!r}'
                    if self.addError(exc.EncodingError(message),
                                     blame='submission',
                                     path=self.path):
                        logd.error(message)

                yield from rows
                return
            except UnicodeDecodeError:
                continue
            except csv.Error as e:
                logd.exception(e)
                message = f'WHAT HAVE YOU DONE {e!r} {self.path.as_posix()!r}'
                if self.addError(message,
                                 blame='EVERYONE',
                                 path=self.path):
                    logd.error(message)

    def xlsx(self):
        try:
            one = list(self.xlsx1())
            e1 = None
        except Exception as e:
            e1 = e
            one = None

        try:
            two = list(self.xlsx2())
            e2 = None
        except Exception as e:
            e2 = e
            two = None

        offset = 2
        if one is not None:
            one_test = [['' if c is None else c for c in r[offset:]] for r in one[offset:]]
        else:
            one_test = None

        if two is not None:
            two_test = [['' if c is None else c for c in r[offset:]] for r in two[offset:]]
        else:
            two_test = None

        #if one_test != two_test:
            #breakpoint()

        if e1 is not None:
            message = f'malformed xlsx file could not be read by xlsx2csv {self.path}'
            if self.addError(message,
                             blame='submission',
                             path=self.path):
                logd.exception(e1)

        if e2 is not None:
            message = f'malformed xlsx file could not be read by openpyxl {self.path}'
            if self.addError(message,
                             blame='submission',
                             path=self.path):
                logd.exception(e2)

        if e2 is None:
            yield from two
        elif e1 is None:
            yield from one
        else:
            raise exc.NoDataError(f'{self.path}') from e2

    def _openpyxl_fixes(self):
        # read sheet read only to find empty columns without destroying memory
        wbro = self._openpyxl.load_workbook(self.path, read_only=True)
        if wbro is None:
            raise exc.NoDataError(f'{self.path}')

        lnn = []
        first_cont_empty = 0
        empty_rows = []
        for i, row in enumerate(wbro.active.rows):
            last_not_none = -1
            for j, cell in enumerate(row):
                if isinstance(cell, self._openpyxl.cell.read_only.EmptyCell) or cell.value is None:
                    pass
                else:
                    last_not_none = j
            lnn.append(last_not_none)
            if last_not_none == -1:  # empty row case
                if not empty_rows or empty_rows and empty_rows[-1] < i - 1:
                    # there was an intervening non-empty row
                    first_cont_empty = i

                empty_rows.append(i)

        # non-contiguous empty rows
        sparse_empty_rows = [e for e in empty_rows if e < first_cont_empty]

        wbro.close()
        s_lnn = set(lnn)
        max_lnn = max(s_lnn) + 1  # + 1 to convert from index -> column number
        start_cols_empty = max_lnn + 1  # start deleting one column after last non-empty
        msg = (f'unique last not nones {s_lnn}, max {max_lnn}, '
               f'delete from {start_cols_empty}')
        log.debug(msg)

        # read sheet rw so that hyperlinks and other stuff work
        wb = self._openpyxl.load_workbook(self.path)
        if wb is None:
            raise exc.NoDataError(f'{self.path}')

        wb.iso_dates=True  # XXX NOTE I think only relevant for writing, useful when run in reverse

        sheet = wb.active
        if sheet.max_column > max_lnn:
            # over count a bit using max_column, but safe if max_lnn
            # is somehow zero
            sheet.delete_cols(start_cols_empty, sheet.max_column)
            msg = f'Empty columns beyond {max_lnn} detected in {self.path}'
            if self.addError(msg,
                             blame='submission',
                             path=self.path):
                logd.warning(msg)

        if first_cont_empty > 0:
            # add 1 since row number starts at 1 not 0
            fce_rn = first_cont_empty + 1
            sheet.delete_rows(fce_rn, sheet.max_row)
            msg = f'Empty rows beyond {first_cont_empty} detected in {self.path}'
            if self.addError(msg,
                             blame='submission',
                             path=self.path):
                logd.warning(msg)

        return sheet, wb, sparse_empty_rows

    def xlsx2(self):
        # FIXME this has horrible performance on sheets with lots of empty cells
        # which apparently happens really frequently in our workflows?!
        # https://openpyxl.readthedocs.io/en/latest/datetime.html  # XXX OH NOOOOOOOOOOOOOOOOOOOOOO
        # yeah ... iso8601 as a string :/

        # xlsx: bringing deep sadness to everyone involved
        sheet, wb, _ser = self._openpyxl_fixes()
        longrs = 0
        emptyrs = 0
        for row in sheet.rows:
            fixed_row = [
                cell.value.date()
                # Who thought it was a good idea to make the internal
                # representation of dates and midnight indistinguishable !?!?
                if cell.value and cell.is_date and
                cell.number_format in ('yyyy\\-mm\\-dd;@', 'YYYY\\-MM\\-DD')
                else (cell.hyperlink.target if cell.hyperlink else cell.value)
                for cell in row]
            # NOTE '@' != 'General' but apparently openpyxl converts @ to number incorrectly
            #breakpoint()
            if not [v for v in fixed_row if v]:
                emptyrs += 1
                continue

            yield fixed_row

        wb.close()
        if emptyrs > 0:
            # should only be catching sparse empty rows at this point
            msg = f'{emptyrs} empty rows detected in {self.path}'
            if self.addError(msg,
                             blame='submission',
                             path=self.path):
                logd.warning(msg)

    def xlsx1(self):
        kwargs = {
            'delimiter': '\t',
            'skip_empty_lines': True,
            'outputencoding': 'utf-8',
            'hyperlinks': True,
        }
        sheetid = 1
        try:
            xlsx2csv = Xlsx2csv(self.path.as_posix(), **kwargs)
        except InvalidXlsxFileException as e:
            raise exc.NoDataError(f'{self.path}') from e

        ns = len(xlsx2csv.workbook.sheets)
        if ns > 1:
            message = f'too many sheets ({ns}) in {self.path.as_posix()!r}'
            if self.addError(exc.EncodingError(message),
                             blame='submission',
                             path=self.path):
                logd.error(message)

        f = io.StringIO()
        try:
            xlsx2csv.convert(f, sheetid)
            f.seek(0)
            gen = csv.reader(f, delimiter='\t')
            yield from gen
        except SheetNotFoundException as e:
            log.warning(f'Sheet weirdness in {self.path}')
            log.warning(str(e))
        except AttributeError as e:
            message = ('xlsx2csv: major sheet weirdness ('
                       'probably a bug in the xlsx2csv converter)? '
                       f'in {self.path}')
            if self.addError(exc.EncodingError(message),
                             blame='submission',
                             path=self.path):
                #log.exception(e)
                logd.warning(message)

    def _bad_filetype(self, type_):
        message = f'bad filetype {type_}\n{self.path.as_posix()!r}'
        raise exc.FileTypeError(message)

    def xls(self):
        return self._bad_filetype('xls')

    def normalize(self, rows):
        # FIXME need to detect changes
        # this removes any columns that are all dead

        #if any(not(c) for c in rows[0]):  # doesn't work because generators

        error = exc.EncodingError(f"encoding feff error in '{self.path}'")
        cleaned_rows = zip(*(t for t in zip(*rows) if not all(not(e) for e in t)))  # TODO check perf here
        for row in cleaned_rows:
            n_row = [c.strip().replace('\ufeff', '') if c and isinstance(c, str) else c
                     for c in row
                     if ((not logd.error(error)
                          if self.addError(error) else
                          True)  # FIXME will probably append multiple ...
                         if c and isinstance(c, str) and '\ufeff' in c else
                         True)]
            if not all(not(c) for c in n_row):  # skip totally empty rows
                yield n_row

    @property
    def normalized(self):
        try:
            yield from self.normalize(getattr(self, self.file_extension)())
        except UnicodeDecodeError as e:
            log.error(f'{self.path.as_posix()!r} {e}')

    @property
    def T(self):
        return self.__class__(self.path, rotate=True)

    def __iter__(self):
        yield from self._iter_(normalize=True)

    def _iter_(self, normalize=False):
        try:
            if normalize:
                try:
                    fef = getattr(self, self.file_extension)
                except AttributeError as e:
                    self._bad_filetype(self.file_extension)

                gen = self.normalize(fef())
            else:
                gen = getattr(self, self.file_extension)()

            if self._rotate:
                yield from zip(*gen)
            else:
                yield from gen
        except UnicodeDecodeError as e:
            log.error(f'{self.path.as_posix()!r} {e}')

    @property
    def title(self):
        path = Path(self.path)
        out = path.name
        if path.cache is not None and path.cache.dataset is not None:
            out += ' ' + path.cache.dataset.name[:30] + ' ...'

        return out

    def __write(self, path, format='csv'):
        # FIXME this isn't actually useful because we already have the file in question
        with open(path, 'wt') as f:
            writer = csv.writer(f)
            writer.writerows(self)

    def __repr__(self):
        limit = 30
        rows = [[c[:limit] + ' ...' if isinstance(c, str)
                 and len(c) > limit else c
                 for c in r] for r in self]
        table = AsciiTable(rows, title=self.title)
        return table.table


ROW_TYPE = object()
COLUMN_TYPE = object()
numberN = object()
GROUP_ALL = object()


def remove_rule(blob, rule):
    # FIXME this should be in _DictTransformer
    previous_target = blob
    target = blob
    previous_element = None
    for path_element in rule:
        if isinstance(path_element, str):
            if path_element in target:
                previous_target = target
                target = previous_target[path_element]
            else:
                return  # target not found

        elif isinstance(path_element, dict):
            if all([(k in target and target[k] == v)
                    for k, v in path_element.items()]):
                previous_target.pop(previous_element)
                return  # we're done here

        else:
            raise ValueError(f'what are you doing with {rule}')

        previous_element = path_element


class PrimaryKey:
    def __init__(self, name, column_names, combine_function,
                 nalt_header, agen, gen):
        self.name = name
        # because we add the primary key to alt header we remove it again here
        # because the rest of the generator values don't have a blank prepended
        nalt_header_less_pk = [nah for nah in nalt_header if nah != name]
        try:
            self.indexes = [nalt_header_less_pk.index(name) for name in column_names]
        except ValueError as e:
            msg = f'Missing a primary key component {e}'
            raise exc.MalformedHeaderError(msg) from e

        self.combine_function = combine_function
        self._agen = agen
        self._gen = gen

    def compute_primary(self, arecord):
        return self.combine_function(tuple(arecord[index] for index in self.indexes))

    def __iter__(self):
        yield (self.name, *[self.compute_primary(ar) for ar in self._agen])
        yield from self._gen

    @property
    def generator(self):
        return iter(self)


class MetadataFile(HasErrors):
    default_record_type = ROW_TYPE
    primary_key_rule = None  # name, the names of the alt_header columns to merge, function to merge tuple
    renames_alt = {}     # renames first since all following
    renames_header = {}  # refs look for the renamed form
    missing_add = {}  # missing also comes after renames
    record_type_key_alt = None  # record type key is affected by renames
    record_type_key_header = None  # record type key is affected by renames
    groups_alt = {}
    groups_header = {}
    ignore_alt = tuple()
    ignore_header = tuple()
    ignore_match = tuple()
    raw_json_class = rj.RawJson
    normalization_class = nml.NormValues  # FIXME normalize SHOULD NOT BE DONE AT THIS STAGE
    normalize_alt = True
    normalize_header = True
    normalize_mismatch_ok = tuple()
    normalize_alt_mismatch_ok = tuple()
    _expect_single = tuple()

    def __new__(cls, path, template_schema_version=None):
        if template_schema_version is not None:  # and template_schema_version < latest  # TODO
            logd.debug(f'{template_schema_version} {path}')  # TODO warn

        if cls.record_type_key_header is None or cls.record_type_key_alt is None:
            raise TypeError(f'record_type_key_? should not be None on {cls.__name__}')

        # TODO check for renames that didn't get picked up
        # especially from the 0, 0 cell

        return super().__new__(cls)

    def __init__(self, path, template_schema_version=None):
        super().__init__()
        self.path = path
        self.template_schema_version = template_schema_version

    def xopen(self):
        self.path.xopen()

    @property
    def data(self):
        self._to_embed = tuple()
        data = self._data()

        condition = False
        if condition:
            breakpoint()

        for embedErrors in self._to_embed:
            embedErrors(data)

        self.embedErrors(data)
        return data

    def _data(self):
        if self.path.suffix == '.json':
            data = self.raw_json_class(self.path).data
            self._expand_string_lists = lambda : data  # FIXME hack
            self.norm_to_orig_header = {}
            def all_keys(d):
                if isinstance(d, dict):
                    for k, v in d.items():
                        if k not in self.groups_alt:
                            yield k

                        yield from all_keys(v)

                elif isinstance(d, list):
                    for e in d:
                        yield from all_keys(e)

            self.norm_to_orig_alt = {k:k for k in all_keys(data)}
            #return self.normalization_class(self).data  # FIXME incredible violation of encapsulation ...
            return self._condense()

        else:
            # TODO if this passes we can just render backward from _normalize
            return self._condense()

    def _condense(self):
        def condense(thing, key=None):
            if isinstance(thing, dict):
                hrm = []
                out = {}
                for k, v in thing.items():
                    nv = condense(v, k)
                    # bah bloody description appearing twice >_<
                    if k in self.norm_to_orig_header and k not in self.norm_to_orig_alt:
                        if is_list_or_tuple(nv):
                            # at this stage there are no lists of lists
                            hrm.extend(nv)
                        else:
                            hrm.append(nv)
                    else:
                        out[k] = nv

                if out and hrm:
                    raise ValueError(f'how!?\n{hrm}\n{out}')
                elif out:
                    return out
                elif hrm:
                    if (key in self._expect_single and
                        len(hrm) == 1):
                        return hrm[0]
                    else:
                        return tuple(hrm)

            #elif is_list_or_tuple(thing):
                #return thing
            else:
                return thing

        data_in = self._cull_rtk()
        data_out = condense(data_in)
        if data_out is None:
            # None breaks further pipelines
            # yes types would help here
            return {}
        else:
            return data_out

    def _cull_rtk(self):
        normalized = self._normalize_values()
        def crtk(thing, only_index_alt=False, only_index_header=False):
            if isinstance(thing, dict):
                out = {}
                for k, v in thing.items():
                    if k in self.groups_alt and self.groups_alt[k] == GROUP_ALL:
                        nv = crtk(v, only_index_alt=True)
                        if nv is not None:
                            out[k] = nv
                    elif k in self.groups_header and self.groups_header[k] == GROUP_ALL:
                        # this isn't used yet but putting it here for symmetry just in case
                        nv = crtk(v, only_index_header=True)
                        if nv is not None:
                            out[k] = nv
                    elif (only_index_alt and k == v == self.record_type_key_alt or
                          only_index_header and k == v == self.record_type_key_header):
                        continue
                    elif (k != self.record_type_key_header and
                          # FIXME requires that headers and alts be fully disjoint
                          k != self.record_type_key_alt or
                          only_index_alt and k == self.record_type_key_alt or
                          only_index_header and k == self.record_type_key_header):
                        nv = crtk(v, only_index_alt, only_index_header)
                        if nv is not None:
                            out[k] = nv

                if out:
                    return out

            else:
                return thing

        crv = crtk(normalized)
        return crv

    def _normalize_values(self):
        nc = self.normalization_class(self)  # calls _normalize_keys internally
        return nc.data

    def _clean(self):
        data_in = self._expand_string_lists()
        def clean(thing):
            if isinstance(thing, dict):
                out = {}
                for k, v in thing.items():
                    cv = clean(v)
                    if cv is not None:
                        out[k] = cv

                if out:
                    return out

            elif thing == '' or thing == {}:
                return None

            elif is_list_or_tuple(thing):
                return [v for v in [clean(v) for v in thing] if v is not None]

            else:
                return thing

        data_out = clean(data_in)
        return data_out

    def _expand_string_lists(self):
        data_in = self._filter()
        # TODO consider whether this step would be useful
        # I think it would because it would vastly simplify
        # the downstream implementation of things like
        # normalizing the contributor roles
        # TODO this might also go before filter?
        # might need to blacklist or whitelist fields?
        seps = '|', ';', ','
        data_out = data_in
        return data_out

    def _filter(self):
        # FIXME does in place modification
        culled = self._cull()
        for rule in self.ignore_match:
            remove_rule(culled, rule)

        return culled

    def _cull(self):
        transformed = self._transformed()
        # FIXME deeper nesting might break this?
        culled = {k:{k:v for k, v in v.items()
                     if k not in self.ignore_header}
                  for k, v in transformed.items()}
        return culled

    def _transformed(self):  # FIXME this whole approach makes it very difficult to test individual stages
        keyed = self._keyed()
        naccounted_baseline = set(e for g in self.groups_alt.values()
                                  if g != GROUP_ALL for e in g)

        # TODO how to flag exclude the header fields ?
        # XXX NOTE failure mode on array but should be string? e.g. study_data_collection
        # the order of operations is why this happens, but the behavior is correct
        # fields that are grouped can and do have multiple entires, so if you are
        # expecting a single object where internally there are multipel entries you
        # you do not get that using the group functionality
        transformed = {}
        for key, alt_grouped_norm in self.groups_alt.items():
            merge = alt_grouped_norm == GROUP_ALL
            if merge:
                alt_grouped_norm = tuple(k for k in self.norm_to_orig_alt
                                         if k not in naccounted_baseline)
                naccounted_baseline.update(set(alt_grouped_norm))

            records = [keyed[nh] for nh in alt_grouped_norm
                       # FIXME versioning ... schema will take care of missing values later
                       if nh in keyed]

            objected = {nh:{record[self.record_type_key_header]:record[nh]
                            for record in records}
                        for nh in self.norm_to_orig_header}

            transformed[key] = {k:v for k, v in objected.items()
                                if v is not None and v != {}}

        nunaccounted = (set(self.norm_to_orig_alt) - naccounted_baseline)

        for nah in nunaccounted: 
            transformed[nah] = keyed[nah]

        return transformed

    def _keyed(self):
        objects = self._add_missing()
        keyed = {o[self.record_type_key_header]:o for o in objects}
        return keyed

    def _add_missing(self):
        objects = self._sanity_check()
        for nah, (value, number) in self.missing_add.items():
            if nah not in self.norm_to_orig_alt:
                self.norm_to_orig_alt[nah] = nah
                cant_go_wrong = {}
                i = 0
                for nh in self.norm_to_orig_header:
                    if ((number == numberN or i < number) and
                        nh != self.record_type_key_header and
                        nh not in self.ignore_header):
                        v = value
                        i += 1
                    else:
                        v = ''

                    cant_go_wrong[nh] = v

                {nh:(value if number == numberN or i <= number else '')
                 for i, nh in enumerate(self.norm_to_orig_header)
                 if nh != self.record_type_key_header}
                #print(cant_go_wrong)
                cant_go_wrong[self.record_type_key_header] = nah
                objects.append(cant_go_wrong)

        # end add missing items
        return objects

    def _sanity_check(self):
        """ first chance we have to check sanity since
            the generators have to have run """
        objects = self._objectify()
        for n, rtk, norm_dict in (('alt_header', self.record_type_key_alt, self.norm_to_orig_alt),
                                  ('header', self.record_type_key_header, self.norm_to_orig_header)):
            if rtk not in norm_dict:
                #breakpoint()  # XXX TODO enable this on debug
                msg = (f'Could not find record primary key for {n} in\n"{self.path}"\n\n'
                       f'{rtk}\nnot in\n{list(norm_dict)}')

                raise exc.MalformedHeaderError(msg)

        return objects

    def _objectify(self):
        gen = self._norm_alt_headers()
        # WARNING assumes dict order
        objects = [{k:v for k, v in zip(self.norm_to_orig_header, r)} for r in gen]
        return objects

    def _norm_alt_headers(self):
        # headers MUST be normalized as early as possible due to the fact that
        # there are very likely to be duplicate keys
        try:
            gen = self._headers()
        except exc.BadDataError as e:
            msg = f'Bad data in {self.path}!'
            raise exc.BadDataError(msg) from e

        indexes = {}
        def normb(k):
            if k in self.orig_to_norm_alt:
                nk = self.orig_to_norm_alt[k]
                if isinstance(nk, tuple):
                    if k in indexes:
                        indexes[k] += 1
                    else:
                        indexes[k] = 0

                    return nk[indexes[k]]  # this is safe because _headers is ordered

            else:
                nk = k

            return nk

        for row in gen:
            yield (normb(row[0]), *row[1:])

    def _headers(self):
        def safe_invert(d):
            out = {}
            for v, k in d.items():
                if k in out:
                    ev = out[k]
                    if isinstance(ev, tuple):
                        out[k] += (v,)
                    else:
                        out[k] = (ev, v)
                else:
                    out[k] = v

            return out

        t = self._t()

        if self.default_record_type == ROW_TYPE:
            alt = t
            head = t.T
        else:
            alt = t.T
            head = t

        self._to_embed = alt.embedErrors, head.embedErrors

        agen = iter(alt)
        gen = iter(head)

        try:
            if self.primary_key_rule is not None:
                pk_name, combine_names, combine_function = self.primary_key_rule
                self.alt_header = (pk_name, *next(agen))
            else:
                self.alt_header = next(agen)
        except StopIteration as e:
            raise exc.NoDataError(f'{self.path}') from e

        self._alt_header = Header(self.alt_header,
                                  normalize=self.normalize_alt,
                                  alt=True,
                                  normalize_first_cell=self.normalize_header)
        self.norm_to_orig_alt = self._alt_header.lut(**self.renames_alt)
        self.orig_to_norm_alt = safe_invert(self.norm_to_orig_alt)
        if self.normalize_header and self.alt_header[0] != self._alt_header.data[0]:
            msg = (f'Potentially bad alt header value {self.alt_header[0]!r} '
                   f'!= {self._alt_header.data[0]!r}'
                   f' in {self.path.as_posix()!r}')
            # this is not guranteed to be an issue but hitting
            # Subject_id instead of subject_id will cause issues for
            # other people who are not using this pipeline to
            # normalize everything (for example)
            if self._alt_header.data[0] not in self.normalize_mismatch_ok:
                logd.debug(msg)

        if self.primary_key_rule is not None:
            nalt_header = [self.orig_to_norm_alt[ah] for ah in self.alt_header]
            gen = PrimaryKey(pk_name, combine_names, combine_function,
                             nalt_header, agen, gen).generator

        # FIXME if the 'header' column is not at position zero
        # this will fail see version 1.1 subjects template for this
        # with a note that 1.1 had many issues
        self.header = next(gen)
        self._header = Header(self.header,
                              normalize=self.normalize_header,
                              normalize_first_cell=self.normalize_alt)
        # FIXME the case where someone has a capitalized subject_id
        self.norm_to_orig_header = self._header.lut(**self.renames_header)
        self.orig_to_norm_header = safe_invert(self.norm_to_orig_header)
        if self.normalize_alt and self.header[0] != self._header.data[0]:
            msg = (f'Potentially bad header value {self.header[0]!r} '
                   f'!= {self._header.data[0]!r}'
                   f' in {self.path.as_posix()!r}')
            if self._header.data[0] not in self.normalize_alt_mismatch_ok:
                logd.debug(msg)

        _matched = set(self.norm_to_orig_alt) & set(self.norm_to_orig_header)
        if _matched:
            if (self.record_type_key_header == self.record_type_key_alt and
                {self.record_type_key_header} == _matched):
                # if 0,0 in the sheet is the key for both with the same name
                # then we skip here because it is expected, we might also add
                # a check that it is in ignored, but I don't think we need it
                pass
            else:
                # this has to be fatal because it violates assumptions made by
                # downstream code that are useful to deal with nesting
                # we may be able to fix this in the future but skip for now
                msg = f'Common cells between alt header and header! {_matched}'
                #log.warning(msg)  # can't quite error yet due to metadata_element
                raise exc.MalformedHeaderError(msg)

        yield self.header
        yield from gen

    def _t(self):
        if self.path.suffix == '.json':
            return  # TODO json -> tabular conversion
        else:
            return Tabular(self.path)


class SubmissionFile(MetadataFile):
    __internal_id_1 = object()
    default_record_type = COLUMN_TYPE
    renames_alt = {'submission_item': __internal_id_1}
    renames_header = {'definition': 'definition_header'}
    record_type_key_alt = __internal_id_1
    record_type_key_header = 'submission_item'
    groups_alt = {'submission': ('sparc_award_number', 'milestone_achieved', 'milestone_completion_date')}
    ignore_alt = __internal_id_1,
    ignore_header = 'definition_header',
    raw_json_class = rj.RawJsonSubmission
    normalization_class = nml.NormSubmissionFile

    @property
    def data(self):
        """ lift list with single element to object """

        data = super().data  # LOL PYTHON can't super in a debugger -- pointless
        d = copy.deepcopy(data)

        if d and 'submission' in d:
            sub = d['submission']
            if sub:
                if is_list_or_tuple(sub):
                    d['submission'] = sub[0]
                elif isinstance(sub, dict):
                    return d
                else:
                    raise TypeError(f'What is a {type(sub)} {sub}?')
            else:
                d['submission'] = {}

        return d


class SubmissionFilePath(ObjectPath):
    obj = SubmissionFile


class CodeDescriptionFile(MetadataFile):
    default_record_type = COLUMN_TYPE
    renames_header = {'description': 'description_header'}
    record_type_key_alt = 'metadata_element'
    record_type_key_header = record_type_key_alt
    ignore_header = 'metadata_element', 'description_header', 'example'


class CodeDescriptionFilePath(ObjectPath):
    obj = CodeDescriptionFile


_props = sc.DatasetDescriptionSchema.schema['properties']
_props2 = sc.ContributorSchema.schema['properties']  # FIXME recurse ...
_nddfes = [k for k, v in chain(_props.items(), _props2.items())
           if isinstance(v, dict) and sc.not_array(v)]
_nddfes = sorted(set(_nddfes))


class DatasetDescriptionFile(MetadataFile):
    default_record_type = COLUMN_TYPE
    renames_alt = {'contributors': 'contributor_name',  # watch out for name collisions (heu heu heu)
                   'type': 'dataset_type',  # type by itself is used internally
                   'metadata_version_do_not_change': 'template_schema_version',
                   'contributor_orcid_id': 'contributor_orcid',  # <= 1.2.3
                   'metadata_version': 'template_schema_version',  # 2.0
                   'acknowledgements': 'acknowledgments',  # <= 1.2.3
                   #'relation_type': 'relation_type', # 2.0
                   'identifier': 'related_identifier',
                   'identifier_description': 'related_identifier_description',
                   'identifier_type': 'related_identifier_type',
                   #'study_organ_system': 'organ',  # XXX do not convert, requires normalization
                   #'study_approach': 'approach',  # XXX do not convert, requires normalization
                   #'study_technique': 'technique',  # XXX do not convert, requires normalization
                   }
    renames_header = {'description': 'description_header'}
    missing_add = {'template_schema_version': ('1.0', 1)}
    record_type_key_alt = 'metadata_element'
    record_type_key_header = record_type_key_alt
    groups_alt = {'contributors': ('contributor_name',
                                   'contributor_orcid',
                                   'contributor_affiliation',
                                   'contributor_role',
                                   'is_contact_person',),
                  'related_identifiers': ('related_identifier',
                                          'related_identifier_type',
                                          'relation_type',
                                          'related_identifier_description',),
                  'study': ('study_purpose',
                            'study_data_collection',
                            'study_primary_conclusion',
                            'study_collection_title',),
                  'links': ('additional_links', 'link_description'),
                  'examples': ('example_image_filename',
                               'example_image_locator',
                               'example_image_description'),}

    ignore_header = 'metadata_element', 'example', 'description_header'
    ignore_alt = ('basic_information', 'study_information', 'contributor_information',
                  'related_protocol__paper__dataset__etc_', 'participant_information',)
    raw_json_class = rj.RawJsonDatasetDescription
    normalization_class = nml.NormDatasetDescriptionFile
    normalize_mismatch_ok = 'metadata_element',
    normalize_alt_mismatch_ok = normalize_mismatch_ok
    _expect_single = _nddfes

    @property
    def data(self):
        data = super().data
        return data


class DatasetDescriptionFilePath(ObjectPath):
    obj = DatasetDescriptionFile


_props = sc.SubjectsSchema.schema['properties']['subjects']['items']['properties']
_props2 = sc.SamplesFileSchema.schema['properties']['samples']['items']['properties']
_props3 = sc._software_schema['items']['properties']
_props4 = sc.PerformancesSchema.schema['properties']['performances']['items']['properties']
_nsffes = [k for k, v in chain(_props.items(), _props2.items(), _props3.items(), _props4.items())
           if isinstance(v, dict) and sc.not_array(v)]
_nsffes = sorted(set(_nsffes))


class PerformancesFile(MetadataFile):
    renames_header = {'performance_id': 'metadata_element',}
    record_type_key_alt = 'performance_id'
    record_type_key_header = 'metadata_element'
    groups_alt = {'performances': GROUP_ALL,}
    normalization_class = nml.NormPerformancesFile
    normalize_header = False
    _expect_single = _nsffes


class PerformancesFilePath(ObjectPath):
    obj = PerformancesFile


class SubjectsFile(MetadataFile):
    #default_record_type = COLUMN_TYPE
    __internal_id_1 = object()
    renames_header = {'subject_id': 'metadata_element',
                      ('Lab-based schema for identifying each subject, '
                       'should match folder names'): __internal_id_1,}
    record_type_key_alt = 'subject_id'
    record_type_key_header = 'metadata_element'
    groups_alt = {'subjects': GROUP_ALL,
                  'software':('software',
                              'software_version',
                              'software_vendor',
                              'software_url',
                              'software_rrid'),}
    ignore_header = __internal_id_1,
    ignore_alt = 'additional_fields_e_g__minds',
    ignore_match = [['subjects', 'sub-1', {'subject_id': 'sub-1',
                                           'pool_id': 'pool-1',
                                           'reference_atlas': 'Paxinos Rat V3',
                                           'handedness': 'right',}]]
    raw_json_class = rj.RawJsonSubjects
    normalization_class = nml.NormSubjectsFile
    normalize_header = False
    _expect_single = _nsffes

    @property
    def data(self):
        data = super().data
        return data


class SubjectsFilePath(ObjectPath):
    obj = SubjectsFile


def join_cast(sep):
    """ oh look, you still can't use list comprehension in class scope """
    def sigh(t, sep=sep):
        for _ in t:
            if _ is None:
                raise exc.BadDataError('A primary key value was None!')

        return sep.join(tuple(str(_) for _ in t))

    return sigh


class SamplesFile(SubjectsFile):
    """ TODO ... """

    # TODO merge
    __internal_id_1 = object()
    __primary_key = 'primary_key'
    primary_key_rule = __primary_key, ('subject_id', 'sample_id'), join_cast('_')
    __rename_1 = primary_key_rule[2](('Lab-based schema for identifying each subject',
                                      ('Lab-based schema for identifying each sample, must '
                                       'be unique')))
    renames_header = {__primary_key: 'metadata_element',
                      __rename_1: __internal_id_1,}
    record_type_key_alt = __primary_key
    record_type_key_header = 'metadata_element'
    groups_alt = {'samples': GROUP_ALL,
                  'software':('software',
                              'software_version',
                              'software_vendor',
                              'software_url',
                              'software_rrid'),}
    ignore_header = __internal_id_1,
    ignore_alt = 'additional_fields_e_g__minds',
    ignore_match = [['samples', primary_key_rule[2](('sub-1', 'sub-1_sam-2')),
                     {'subject_id': 'sub-1',
                      'pool_id': 'pool-1',
                      'reference_atlas': 'Paxinos Rat V3',
                      'handedness': 'right',}]]
    normalization_class = nml.NormSamplesFile
    normalize_header = False
    _expect_single = _nsffes


class SamplesFilePath(ObjectPath):
    obj = SamplesFile


# TODO when we need them ?
#_props = sc.ManifestFileSchema.schema['properties']['man']['items']['properties']
#_nsman = [k for k, v in _props.items()
          #if isinstance(v, dict) and sc.not_array(v)]
#_nsman = sorted(set(_nsffes))
class ManifestFile(MetadataFile):  # FIXME need a PatternManifestFile I think?
    renames_header = {'filename': 'metadata_element',}
    record_type_key_alt = 'filename'
    record_type_key_header = 'metadata_element'
    groups_alt = {'manifest_records': GROUP_ALL}
    normalize_header = False
    #raw_json_class = rj.RawJsonManifestFile  # FIXME TODO sigh
    #_expect_single = _nsman

    @staticmethod
    def fix_tod_123(blob):
        if 'timestamp' in blob and 'timestamp_or_date' not in blob:
            # XXX FIXME this destroys the original view of the data
            blob['timestamp_or_date'] = blob.pop('timestamp')

    @property
    def data(self):
        try:
            data = super().data
        except exc.MalformedHeaderError as e:
            if (type(self) == ManifestFile and
                (hasattr(self, 'norm_to_orig_header') and
                 'pattern' in self.norm_to_orig_header and
                 'filename' not in self.norm_to_orig_header or
                 # try to see if pattern is there, but if we have to
                 # run this branch then we are already in trouble
                 'pattern' in self.header and
                 'pattern' in self.alt_header and
                 'filename' not in self.header)):
                log.info(e)  # not much point in logging the full exception
                # FIXME massive hack to work around the multiple
                # different schemas
                return PatternManifestFile(
                    self.path, self.template_schema_version).data
            else:
                raise e

        if isinstance(self.template_schema_version, tuple):
            # error will be caught via schema and will also lead to a
            # multiplication of errors since the elif here won't run
            pass
        elif self.template_schema_version is None:
            # this can happen on a NoDataError
            pass
        elif self.template_schema_version <= '1.2.3': # FIXME HACK
            if 'manifest_records' in data:  # raw json manifests don't conform to this structure FIXME
                [self.fix_tod_123(blob) for blob in data['manifest_records']]

        return data


class PatternManifestFile(ManifestFile):
    # FIXME this is an extreme hack and we need to resolve how we are
    # going to deal with with pattern vs filename because right now
    # you can't really have two valid schemas under a single name,
    # though in principle there is no issue with that
    renames_header = {'pattern': 'metadata_element',}
    record_type_key_alt = 'pattern'
    record_type_key_header = 'metadata_element'
    groups_alt = {'manifest_records': GROUP_ALL}
    normalize_header = False


class ManifestFilePath(ObjectPath):
    obj = ManifestFile


# XXX NOTE *_file is added to all of these keys in DatasetStructure.data
DatasetStructure.sections = {'submission': SubmissionFilePath,
                             'code_description': CodeDescriptionFilePath,
                             'dataset_description': DatasetDescriptionFilePath,
                             'performances': PerformancesFilePath,
                             'subjects': SubjectsFilePath,
                             'samples': SamplesFilePath,
                             'manifest': ManifestFilePath,}

DatasetStructure._bind_sections()
DatasetStructure._pipeline_stage = DatasetStructure.__name__



import itertools
from functools import wraps
from collections import defaultdict
from urllib.parse import quote
from datetime import datetime, date
import idlib
from pyontutils.utils import isoformat
from pyontutils.sheets import Sheet as SheetBase, get_cells_from_grid
from sparcur import exceptions as exc
from sparcur.core import adops, OntId, OntTerm
from sparcur.utils import log, logd
from sparcur.config import auth

# google sheets

class Sheet(SheetBase):

    _do_cache = False  # retrieve to cache
    _re_cache = False  # refresh the cache
    _only_cache = False  # cache only aka no network

    def __init__(self, *args, **kwargs):
        self._cache_path = auth.get_path('cache-path') / 'google_sheets'
        if not self._only_cache:
            self._setup_saf(**kwargs)

        try:
            super().__init__(*args, **kwargs)
        finally:
            self._saf = None

    def _setup_saf(self, **kwargs):
        try:
            if 'readonly' not in kwargs or kwargs['readonly']:
                # readonly=True is default so we take this branch if not set
                self._saf = auth.get_path(
                    'google-api-service-account-file-readonly')
            else:
                self._saf = auth.get_path(
                    'google-api-service-account-file-rw')
        except KeyError as e:
            log.warn(e)
        except Exception as e:
            log.exception(e)

    def _setup(self):
        if not self._only_cache:
            super()._setup()

    def fetch(self, _refresh_cache=False, **kwargs):
        if self._do_cache or self._only_cache:
            cache = idlib.cache.cache(self._cache_path, create=True, return_path=True)
            @cache
            def cache_wrap_fetch(name, sheet_name, __f=super().fetch):
                return __f(**kwargs)

            cache_wrapper = cache_wrap_fetch
            (meta, meta_file, raw_values, raw_values_formula, grid), path = cache_wrapper(
                self.name, self.sheet_name,
                _refresh_cache=_refresh_cache or self._re_cache)
            values = [list(r) for r in  # XXX FIXME code duplication
                      zip(*itertools.zip_longest(*raw_values,
                                                 fillvalue=''))]
            values_formula = [list(r) for r in
                              zip(*itertools.zip_longest(*raw_values_formula,
                                                         fillvalue=''))]

            if grid:
                # XXX FIXME duplicated code from pyontutils sheets
                filter_cell = None
                cells = get_cells_from_grid(grid, self.sheet_name, filter_cell)
                cells_index = {(i, j):v for i, j, v in cells}
            else:
                cells_index = {}

            self._meta = meta
            self._meta_file = meta_file
            self.raw_values = raw_values
            self._values = values
            self.raw_values_formula = raw_values_formula
            self._values_formula = values_formula
            self.grid = grid
            self.cells_index = cells_index
            return meta, meta_file, raw_values, values, grid

        return super().fetch(**kwargs)

    def metadata(self, _refresh_cache=False):
        if self._do_cache or self._only_cache:
            cache = idlib.cache.cache(self._cache_path, create=True, return_path=True)
            @cache
            def cache_wrap_meta(name, sheet_name, __f=super().metadata):
                return __f()

            cache_wrapper = cache_wrap_meta
            meta, path = cache_wrapper(
                self.name, self.sheet_name,
                _refresh_cache=_refresh_cache or self._re_cache)
            self._meta = meta
            return meta

        return super().metadata()

    def metadata_file(self, _refresh_cache=False):
        if self._do_cache or self._only_cache:
            cache = idlib.cache.cache(self._cache_path, create=True, return_path=True)
            @cache
            def cache_wrap_meta_file(name, sheet_name, __f=super().metadata_file):
                return __f()

            cache_wrapper = cache_wrap_meta_file
            meta_file, path = cache_wrapper(
                self.name, self.sheet_name,
                _refresh_cache=_refresh_cache or self._re_cache)
            self._meta_file = meta_file
            return meta_file

        return super().metadata_file()


# master
class Master(Sheet):
    name = 'sparc-master'


class Overview(Master):
    sheet_name = 'July18_2019'

    def dataset_ids(self):
        header, *rows = self.values
        index = header.index('identifier')
        return set(i for i in [row[index]for row in rows] if i)


class Progress(Master):
    sheet_name = 'Curation Progess (OT Only)'


class Grants(Master):
    sheet_name = 'Grant to Blackfynn to Protocols.io CLEAN'


class ISAN(Master):
    sheet_name = 'ISAN Demo July 2019'


class Participants(Master):
    sheet_name = 'Participan Notes'


class Protocols(Master):
    sheet_name = 'Protocol URL--> Blackfynn URL'


# affiliations
class Affiliations(Sheet):
    name = 'sparc-affiliations'
    sheet_name = 'Affiliations'
    index_columns = 'affiliation_string',

    def __call__(self, affiliation_string):
        if not isinstance(affiliation_string, str):
            logd.critical(str(affiliation_string))
            return self(affiliation_string[0] + 'ERROR ERROR')

        m = self.mapping

        if not isinstance(affiliation_string, str):
            log.critical('sigh')
            return None

        if affiliation_string in m:
            return m[affiliation_string]
        else:
            # FIXME super inefficient
            las = len(affiliation_string)
            for l, s in sorted([(len(k), k) for k in m.keys()], reverse=True):
                if l <= las and s in affiliation_string:
                    return m[s]

    @property
    def mapping(self):
        def ror_id(row):
            r = row.ror_id().value
            return idlib.Ror(r) if r else None

        return {row.affiliation_string().value:ror_id(row) for row in self.rows()[1:]}


# class sparc reports

class Reports(Sheet):
    """ spreadsheet that contains all spc reports as tabs """
    name = 'spc-reports'

    @classmethod
    def asPreviewClass(cls):
        class ReportPreviewClass(ReportsPreview, cls):
            pass

        assert ReportPreviewClass.name != cls.name
        assert ReportPreviewClass.name.startswith(cls.name), (ReportPreviewClass.name, cls.name)
        return ReportPreviewClass

    @classmethod
    def makeReportSheet(cls, *index_cols, sheet_name=None):
        """ Decorator to build a Sheet class for a cli.Report method

            NOTE reports that do not have index columns will update
            the entire sheet every time, so probabaly don't edit those
        """
        _sheet_name = sheet_name  # avoid issues with class scope
        del sheet_name

        def _rowcellify(cell_value):
            if isinstance(cell_value, idlib.Stream):
                try:
                    return cell_value.asUri()
                except AttributeError as e:
                    breakpoint()
            elif (isinstance(cell_value, date) or
                  isinstance(cell_value, datetime)):
                # the default json encoder doesn't know what to do with these
                return isoformat(cell_value)
            elif isinstance(cell_value, str):
                if cell_value.lower() == 'x':
                    return True

            return cell_value

        def rowcellify(rows):
            return [[_rowcellify(cell_value) for cell_value in columns]
                    for columns in rows]

        def decorator(method):
            tags = method.__name__ == 'anno_tags'

            class ReportSheet(cls):
                __name__ = f'Report{method.__name__.capitalize()}'
                sheet_name = method.__name__ if _sheet_name is None else _sheet_name
                index_columns = index_cols

            method._report_class = ReportSheet

            @wraps(method)
            def inner(self, *args, **kwargs):
                if self.options.preview:
                    RC = method._report_class.asPreviewClass()
                else:
                    RC = method._report_class

                if 'ext' not in kwargs or kwargs['ext'] is None:
                    kwargs['ext'] = True
                elif isinstance(kwargs['ext'], bool) and kwargs['ext']:
                    pass  # already went through
                elif self.options.to_sheets:
                    msg = 'Should not have extension and export to sheets.'
                    raise TypeError(msg)
                else:
                    pass

                out = method(self, *args, **kwargs)

                if self.options.to_sheets:
                    if tags:
                        if args:
                            RC.sheet_name = args[0]
                        else:
                            RC.sheet_name = self.options.tag[0]

                    rows, title = out
                    # fetch=False since the remote sheet may not exist
                    # TODO improve the ergonimcs here
                    report = RC(fetch=False, readonly=False)
                    report.metadata()  # idlib.Stream vs ? behavior not decided
                    rows_stringified = rowcellify(rows)  # FIXME align to upstream schema
                    if report.sheetId() is None:
                        report.createRemoteSheet()
                        report.update(rows_stringified)
                    else:
                        report.fetch()
                        if report.index_columns and report.values:
                            # FIXME bad assumption that column
                            # ordering remains the same?  FIXME upsert
                            # should be ok in this situation but
                            # dropping the header causes issues
                            # XXX sort of fixedish no longer mismatching now at least
                            remap = [rows[0].index(c) if c in rows[0] else None
                                     for c in report.values[0]]
                            aligned = [[None if i is None else rs[i]
                                        for ui, i in enumerate(remap)]
                                       for rs in rows_stringified[1:]]
                            def gmr(a):
                                try:
                                    return report._row_from_index(row=a)[0].values
                                except AttributeError:
                                    return a

                            preserved = [[cr if ca is None else ca for cr, ca in zip(r, a)]
                                         for r, a in zip([gmr(a) for a in aligned], aligned)]
                            # FIXME seems to produce duplicate rows or something?
                            report.upsert(*preserved)
                        else:
                            report.update(rows)

                    report.commit()

                return out

            return inner

        return decorator


class ReportsPreview(Reports):
    """ reporting for the preview of the next iteration of the pipelines """
    name = 'spc-reports-preview'


class Mis(Reports):
    sheet_name = 'mis'

    # NOTE a ReportSheet for this sheet is generated as well
    # it is simple write-only functionality, so we create
    # another one here instead of hacking around
    # reports.Report.mis._report_class

    def triples(self):
        from pyontutils import namespaces as ns
        import rdflib
        for row in self.rows()[1:]:
            yield from self._rowhack(row, ns, rdflib)

    @staticmethod
    def _rowhack(self, ns, rdflib):
        # can't overwrite pyontutils.sheets.Row like we do in other cases sigh
        curie = self.curie().value
        if not curie:
            return

        oid = OntId(curie)
        s = oid.u

        _lon = (lambda v: None if not v else
                rdflib.Literal(v))
        lon = lambda c: _lon(c.value)

        _blon = lambda v: _lon(True if v == 'TRUE' else False)  # sigh sheets datatypes
        blon = lambda c: _blon(c.value)

        _oon = (lambda v: None if not v else
                OntId(v).u)
        oon = lambda c: _oon(c.value)

        self.type().value
        self.mistake().value
        pos = (
            (ns.ilxtr.curationInternal, blon(self.curation_internal())),
            (ns.definition, lon(self.definition())),
            (ns.editorNote, lon(self.notes())),
            (ns.rdf.type, oon(self.rdf_type())),
            (ns.replacedBy, oon(self.replacedby_())),
            (ns.ilxtr.futureType, oon(self.future_type())),
        )

        for p, o in pos:
            if o is not None:
                yield s, p, o


class AnnoTags(Reports):
    name = 'anno-tags'

    @classmethod
    def asPreviewClass(cls):
        """ anno tags don't need a preview class """
        return cls

    def _annotation_row(self, anno, edfix=False):
        key = self.index_columns[0]
        value = getattr(anno, key)
        value = value.strip()  # FIXME align with id normalization maybe?
        if edfix and value.endswith('ed'):
            ree = None
            for i in range(3):
                try:
                    return self._kv_row(key, value)
                except exc.NotMappedError as e:
                    if i == 0:
                        ree = e
                        value = value[:-2]
                    elif i == 1:
                        value = value + 'e'
                    else:
                        raise ree
        else:
            return self._kv_row(key, value)

    def _kv_row(self, key, value):
        header = self.row_object(0)
        #index = getattr(header, key)().column  # ;_; but it was so elegant
        #row = getattr(index, value)().row
        if key in header._trouble:
            index = header._trouble[key]().column
        else:
            index = getattr(header, key)().column

        if value in index._trouble:
            row = index._trouble[value]().row
        else:
            try:
                row = getattr(index, value)().row
            except AttributeError as e:
                try:
                    # FIXME yes indeed the obscured rules for
                    # when to downcase have come be to haunt us
                    # this was filtering out countless potential matches
                    row = getattr(index, value.lower())().row
                except AttributeError as e:
                    # TODO
                    raise exc.NotMappedError(f'no record for {value}') from e

        return row


class WorkingExecVerb(AnnoTags):
    sheet_name = 'working-protc:executor-verb'
    index_columns = 'value',

    def condense(self):
        marked_as_done = '-'
        mapping = defaultdict(list)

        def make_key(row):
            return tuple(c.value for c in [row.tag(), row.value(), row.text(), row.exact()])

        create = []
        for mt_cell in self.row_object(0).map_to().column.cells[1:]:
            if (mt_cell.value and mt_cell.value != marked_as_done and
                mt_cell.value != mt_cell.row.value().value):
                try:
                    row, iv = self._row_from_index(value=mt_cell.value)
                    key = make_key(row)
                except AttributeError as e:  # value not in index
                    key = ('protc:executor-verb', mt_cell.value, '', '')
                    if key not in create:
                        create.append(key)
                        log.exception(e)

                mapping[key].append(mt_cell.row.value().value)  # cells don't move so we're ok

        mapping = dict(mapping)
        value_to_map_to = {value:k for k, values in mapping.items()
                           for value in values}

        #breakpoint()
        return value_to_map_to, create  # old -> new, original -> correct

    def map(self, anno):
        row = self._annotation_row(anno, edfix=True)
        bad = row.bad().value
        help__ = row.help__().value
        map_to = row.map_to().value
        merge = row.merge().value

        if not (bad or help__):
            if map_to:
                if map_to == '-':
                    # suffix = getattr(anno, self.index_columns[0])  # XXX ignores normalization
                    # XXX TODO consider moving this kind of normalization to the anno class
                    # to avoid these situations
                    suffix = row.value().value
                else:
                    suffix = map_to
            elif merge:
                suffix = merge
            else:
                return  # FIXME error in this case?

            #if suffix == 'anaesthetize':  # finally, was a case conversion issue
                #breakpoint()
            return 'verb:' + suffix.replace(' ', '-')


class WorkingTechniques(AnnoTags):
    sheet_name = 'working-ilxtr:technique'
    index_columns = 'exact',  # FIXME pull index_columns the reports sheets decorator somehow?

    def map(self, anno):
        row = self._annotation_row(anno)
        oid = row.ontology_id().value
        label = row.ontology_label().value
        ilx_curie = row.interlex_id().value
        return OntTerm(oid, label=label), OntTerm(interlex_curie)


class WorkingAspects(AnnoTags):
    sheet_name = 'working-protc:aspect'
    index_columns = 'value',

    def map(self, anno):
        row = self._annotation_row(anno)
        bad = row.bad_().value
        notes_help = row.notes_help().value
        aspect = row.aspect().value
        parent_aspect = row.parent_aspect().value

        if not (bad or notes_help):
            if aspect:
                return 'asp:' + aspect.replace(' ', '-')
            elif parent_aspect:
                return 'asp:' + parent_aspect.replace(' ', '-')


class WorkingAspectsImplied(AnnoTags):
    sheet_name = 'working-protc:implied-aspect'
    index_columns = 'value',

    def map(self, anno):
        row = self._annotation_row(anno)
        map_to = row.map_to().value
        bad = row.bad().value
        if not bad and map_to:
            if map_to == '-':
                suffix = getattr(anno, self.index_columns[0])
            else:
                suffix = map_to

            return 'asp:' + suffix.replace(' ', '-')


class WorkingBlackBox(AnnoTags):
    sheet_name = 'working-protc:black-box'
    index_columns = 'value',

    def map(self, anno):
        row = self._annotation_row(anno)
        mapping_ok = row.mapping_ok().value == 'TRUE'  # FIXME
        not_input = row.not_input_().value
        bad_for_mapping = row.bad_for_mapping_().value
        manual_mapping = row.manual_mapping().value.split('|')[0]
        if mapping_ok and not not_input:
            pass

        if manual_mapping and ' ' in manual_mapping:
            log.error(f'Why does a manual mapping have a space in it {manual_mapping!r}')

        elif manual_mapping:
            return OntTerm(manual_mapping)

        elif mapping_ok:  # FIXME anno.astValue can drift from auto_mapping
            # this is so hilariously inefficient, we parse the same stuff
            # 3 times or something
            _v = anno.asPython().asPython()
            if hasattr(_v, 'black_box'):  # black-box-components only have a name
                # XXX this will fail when black_box is not a term because
                # mapping_ok is true but we removed the old mapping
                if hasattr(_v.black_box, 'curie'):
                    return OntTerm(_v.black_box.curie)
                else:
                    raise exc.NotMappedError(f'no curie mapping {_v}')
            else:
                raise exc.NotMappedError(f'no black_box mapping {_v} ')


class WorkingBlackBoxComponent(WorkingBlackBox):
    sheet_name = 'working-protc:black-box-component'


class WorkingInputs(WorkingBlackBox):
    sheet_name = 'working-protc:input'


class WorkingInputInstances(AnnoTags):
    sheet_name = 'working-protc:input-instance'
    index_columns = 'exact',

    def map(self, anno):
        row = self._annotation_row(anno)
        proposed_category = row.proposed_category().value
        interlex = row.interlex().value
        rrid = row.rrid().value
        # TODO multiple returns
        if rrid:
            return rrid


# field alignment
class FieldAlignment(Sheet):
    name = 'sparc-field-alignment'


class Subjects(FieldAlignment):
    sheet_name = 'Subjects'


class Keywords(FieldAlignment):
    sheet_name = 'All Keywords'


class KeywordSets(FieldAlignment):
    sheet_name = 'Keyword sets'


class _Organs(FieldAlignment):
    # just having fetch_grid = True is enough to bolox everything
    sheet_name = 'Organs'
    index_columns = 'id',
    fetch_grid = True  # THIS IS THE CULPRET
    #fetch_grid = False  # THIS FLIES
    def modality(self, dataset_id): None
    def organ_term(self, dataset_id): None
    def award_manual(self, dataset_id): None
    def techniques(self, dataset_id): return []
    def protocol_uris(self, dataset_id): return []


class Organs(FieldAlignment):
    sheet_name = 'Organs'
    index_columns = 'id',
    fetch_grid = False  # only need hyperlinks, now via FORMULA so no grid
    #fetch_grid = True
    _news = []

    def _lookup(self, dataset_id, fail=False, raw=True):
        try:
            row, iv = self._row_from_index('id', dataset_id)
            return row
        except AttributeError as e:
            # TODO update the sheet automatically
            if dataset_id not in self._news:
                log.critical(f'New dataset! {dataset_id}')
                self._news.append(dataset_id)

            if fail:
                raise e

    def _dataset_row_index(self, dataset_id):
        # FIXME dict or cache this for performance
        row = self._lookup(dataset_id, fail=True)
        return self.values.index(row.values)

    def _update_technique(self, cell):
        # NOTE some rows won't update if the dataset no longer exists
        value = cell.value
        if value:
            try:
                term = next(OntTerm.query(label=value))
                cell.value = term.asCellHyperlink()
            except StopIteration:
                log.info(f'no term for technique {value}')

    def _update_dataset_metadata(self, *, id, name, award, species, update_techniques=False):
        try:
            row_index = self._dataset_row_index(id)
            row = self.row_object(row_index)

            if update_techniques:
                for i in range(1,10):
                    h = f'technique{i}'
                    cell_t = getattr(row, h, lambda:None)()
                    if cell_t is not None:
                        self._update_technique(cell_t)

            cell_dsn = row.dataset_name()
            if cell_dsn.value != name:
                cell_dsn.value = name

            cell_award = row.award()
            if cell_award.value != award:
                cell_award.value = award

            cell_species = row.species()
            if cell_species.value != species:
                cell_species.value = species

        except AttributeError as e:
            row = ['', '', name, id, award, species]  # XXX this was causing index errors
            # because the columns were not being padded by appendRow
            self._appendRow(row)

    def update_from_ir(self, ir):
        oi = OntTerm.query._instrumented
        if oi is not OntTerm:
            OntTerm.query._instrumented = OntTerm

        def cformat(cell):
            if isinstance(cell, OntTerm):
                cell = cell.asCell()

            return cell

        try:
            dataset_blobs = ir['datasets']
            self._wat = self.values[8]
            for blob in dataset_blobs:
                meta = blob['meta']
                #species = adops.get(blob, ['subjects', int, 'species'], on_failure='')  # TODO not implemented
                if 'subjects' in blob:
                    species = '\n'.join(sorted(set(
                        [cformat(s['species']) for s in blob['subjects']
                         if 'species' in s])))
                else:
                    species = ''

                self._update_dataset_metadata(
                    id=blob['id'],
                    name=adops.get(blob, ['meta', 'folder_name'], on_failure=''),
                    award=adops.get(blob, ['meta', 'award_number'], on_failure=''),
                    species=species,
                )
        finally:
            # FIXME this is so dumb :/
            OntTerm.query._instrumented = oi
        log.debug(self.uncommitted())
        self.commit()

    def modality(self, dataset_id):
        """ tuple of modalities """
        row = self._lookup(dataset_id)
        if row:
            try:
                m1 = row.modality1()
                m2 = row.modality2()
                m3 = row.modality3()
                return tuple(_.value for _ in (m1, m2, m3) if _.value)
            except AttributeError as e:
                raise ValueError(f'issue in {self.name} {self.sheet_name}') from e

    def organ_term(self, dataset_id):
        row = self._lookup(dataset_id)
        if row:
            organ_term = row.organ_term()
            otv = organ_term.value
            ot = otv if otv else None
            if ot:
                try:
                    ts = tuple(OntId(t) for t in ot.split(' ') if t and t.lower() != 'na')
                    return ts
                except OntId.BadCurieError:
                    log.error(ot)

    def award(self, dataset_id):
        row = self._lookup(dataset_id)
        if row:
            award = row.award()
            av = award.value
            award_manual = row.award_manual()
            amv = award_manual.value
            return av if av else (amv if amv else None)
        
    def award_machine(self, dataset_id):
        row = self._lookup(dataset_id)
        if row:
            award = row.award()
            av = award.value
            return av if av else None

    def award_manual(self, dataset_id):
        row = self._lookup(dataset_id)
        if row:
            award_manual = row.award_manual()
            amv = award_manual.value
            return amv if amv else None

    def techniques(self, dataset_id):
        row = self._lookup(dataset_id)

        def mkval(cell):
            hl = cell.hyperlink
            if hl is not None:
                oid = OntId(hl)
                if oid.prefix == 'TEMP':
                    logd.warning(f'{cell.value} -> {oid!r}')
                    #return OntTerm(curie=f'lex:{quote(cell.value)}')
                #else:

                return oid.asTerm()

            else:
                logd.warning(f'unhandled technique {cell.value}')
                return cell.value

        if row:
            return [mkval(c) for d in dir(row)
                    if d.startswith('technique')
                    for c in (getattr(row, d)(),)
                    if c.value and c.value.lower() != 'na']
        else:
            return []

    def protocol_uris(self, dataset_id):
        row = self._lookup(dataset_id)

        def mkval(cell):
            hl = cell.hyperlink
            cv = cell.value
            if hl is None:
                hl = cv if cv else None

            if hl is not None:
                try:
                    return idlib.Pio(hl)
                except idlib.exc.IdlibError as e:
                    try:
                        return idlib.Doi(hl)
                    except idlib.exc.IdlibError as e:
                        # XXX WARNING the fall through here
                        # kicks the can down the road
                        pass

            logd.warning(f'unhandled value {cell.value!r}')
            return cv

        if row:
            skip = '"none"', 'NA', 'na', 'no protocols', 'take protocol from other spreadsheet, '
            return [mkval(c) for d in dir(row)
                    if d.startswith('protocol')
                    for c in (getattr(row, d)(),)
                    if c.value and c.value not in skip]
        else:
            return []


def mkval_tech(cell):
    hl = cell.hyperlink
    if hl is not None:
        oid = OntId(hl)
        if oid.prefix == 'TEMP':
            logd.warning(f'{cell.value} -> {oid!r}')
            #return OntTerm(curie=f'lex:{quote(cell.value)}')
        #else:

        return oid.asTerm()

    else:
        logd.warning(f'unhandled technique {cell.value}')
        return cell.value


def main():
    # simple testing
    o = Organs()
    print(o.values[0])


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# [[file:../../docs/developer-guide.org::export.py][export.py]]
import json
from socket import gethostname
import augpathlib as aug
from pyontutils.utils import timeformat_friendly
import sparcur
from sparcur.core import JEncode
from sparcur.paths import Path
from sparcur.utils import loge, symlink_latest
from sparcur.config import auth


def prepare_dataset_export(path, export_path=None):  # FIXME do we need export_base?
    if export_path is None:  # FIXME confusing and breaks w/ convention -> Options maybe?
        export_path = Path(auth.get_path('export-path'))  # FIXME allow alt?

    from sparcur.utils import PennsieveId
    identifier = PennsieveId(path.cache.id)
    uuid = identifier.uuid
    # we don't use cache._fs_safe_id here because we know the
    # identifier type from the folder structure
    # FIXME dataset metadata export setup basically needs to do all of this first
    # set latest run and then latest complete at the end, but complete is kind of arbitrary
    # from the atomic point of view
    tupdated = path.updated_cache_transitive()  # FIXME this causes us to traverse all files twice
    # XXX TODO I think that the dataset updated date is now transitive as well? though the dataset
    # updated timestamp seems to happen a bit before the created/updated date on the file itself?
    # FIXME somehow tupdated can be None !??!?! XXX yes, it happens on empty sparse datasets
    export_dataset_folder = export_path / 'datasets' / uuid
    export_parent = export_dataset_folder / timeformat_friendly(tupdated)
    if not export_parent.exists():
        export_parent.mkdir(parents=True)

    export_latest_run = export_dataset_folder / 'LATEST_RUN'
    symlink_latest(export_parent, export_latest_run)
    # FIXME need to symlink to LATEST
    return export_parent


def export_blob(blob, blob_file_name, time_now=None,
                export_path=None, export_parent_path=None, **kwargs):
    if export_parent_path is None:
        # NOTE if export_parent_path is not None then it is up to the user
        # to make sure that the contents of the dataset directory do not change
        # resulting in confusion about mismatched transitive update dates
        export_parent_path = prepare_dataset_export(path, export_path)
    elif not export_parent_path.exists():
        # safe to mkdir here since outside has a record of the path name
        export_parent_path.mkdir(parents=True)

    export_blob_path = export_parent_path / blob_file_name

    add_prov(blob, time_now)

    with open(export_blob_path, 'wt') as f:
        json.dump(blob, f, indent=2, cls=JEncode)

    loge.info(f'path metadata exported to {export_blob_path}')
    return export_blob_path


def add_prov(blob, time_now):
    """ Mutate blob in place to add prov. """
    # FIXME this will klobber cases where prov was embedded by the pipelines
    blob['prov'] = {'timestamp_export_start': time_now.START_TIMESTAMP,
                    'export_system_identifier': Path.sysid,
                    'export_hostname': gethostname(),
                    'sparcur_version': sparcur.__version__,
                    'sparcur_internal_version': sparcur.__internal_version__,
                    }
    rp = aug.RepoPath(sparcur.core.__file__)
    if rp.working_dir is not None:
        blob['prov']['sparcur_commit'] = rp.repo.active_branch.commit.hexsha


def main(path=Path.cwd(), export_path=None, **kwargs):
    return prepare_dataset_export(path, export_path=export_path)


if __name__ == '__main__':
    import pprint
    from sparcur.simple.utils import pipe_main
    pipe_main(main, after=pprint.pprint)
# export.py ends here



#!/usr/bin/env python3
# [[file:../../docs/developer-guide.org::fetch_remote_metadata.py][fetch_remote_metadata.py]]
from joblib import Parallel, delayed
from sparcur.paths import Path
from sparcur.backends import PennsieveDatasetData


def _from_project_path_fetch_remote_metadata(project_path, n_jobs=12, cached_ok=False):
    if n_jobs <= 1:
        prepared = [PennsieveDatasetData(dataset_path.cache)
                    for dataset_path in project_path.children]
        [bdd() for bdd in prepared if not (cached_ok and bdd.cache_path.exists())]
    else:
        fetch = lambda bdd: bdd() if not (cached_ok and bdd.cache_path.exists()) else None
        fetch_path = (lambda path: fetch(PennsieveDatasetData(path.cache)))
        Parallel(n_jobs=n_jobs)(delayed(fetch_path)(dataset_path)
                 for dataset_path in project_path.children)


# fetch remote metadata
def from_path_fetch_remote_metadata(path, n_jobs=12, cached_ok=False):
    """ Given a path fetch remote metadata associated with that path. """

    cache = path.cache
    if cache.is_organization():
        _from_project_path_fetch_remote_metadata(path, n_jobs=n_jobs, cached_ok=cached_ok)
    else:  # dataset_path
        # TODO more granular rather than roll up to dataset if inside?
        bdd = PennsieveDatasetData(cache)
        if not (cached_ok and bdd.cache_path.exists()):
            bdd()


def main(path=Path.cwd(), n_jobs=12, rmeta_cached_ok=False, **kwargs):
    if path is None or path.find_cache_root() not in (path, *path.parents):
        from sparcur.simple.clone import main as clone
        path = clone(path=path, n_jobs=n_jobs, **kwargs)
        # NOTE path is passed along here, but kwargs is expected to contain
        # parent_path or parent_parent_path and project_id note that if that
        # happens then the path returned from clone will change accordingly

    from_path_fetch_remote_metadata(path, n_jobs=n_jobs, cached_ok=rmeta_cached_ok)
    return path


if __name__ == '__main__':
    from sparcur.simple.utils import pipe_main
    pipe_main(main)  # we probably don't print here?
# fetch_remote_metadata.py ends here



#!/usr/bin/env python3
# [[file:../../docs/developer-guide.org::fetch_remote_metadata_all.py][fetch_remote_metadata_all.py]]
from joblib import Parallel, delayed
from sparcur.backends import PennsieveDatasetData
from sparcur.simple.utils import backend_pennsieve


def from_id_fetch_remote_metadata(id, project_id=None, n_jobs=12):
    """ given an dataset id fetch its associated dataset metadata """
    if id.type == 'organization':
        RemotePath = backend_pennsieve()
        project = RemotePath(id)
        prepared = [PennsieveDatasetData(r) for r in project.children]
        if n_jobs <= 1:
            [p() for p in prepared]
        else:
            # FIXME Paralle isn't really parallel here ...
            # can't use multiprocessing due to broken aug.RemotePath implementation
            # LOL PYTHON everything is an object, except when you want to pickle it
            # then some objects are more equal than others
            Parallel(n_jobs=n_jobs)(delayed(p._no_return)() for p in prepared)

    elif id.type == 'dataset':
        RemotePath = backend_pennsieve(project_id)
        dataset = RemotePath(id)
        bdd = PennsieveDatasetData(dataset)
        bdd()
    else:
        raise NotImplementedError(id)


def main(id=None,
         project_id=None,
         project_id_auth_var='remote-organization',  # FIXME move to clifun
         n_jobs=12,
         **kwargs):

    if project_id is None:
        from sparcur.utils import PennsieveId
        from sparcur.config import auth
        project_id = auth.get(project_id_auth_var)
        project_id = PennsieveId(project_id)  # FIXME abstract the hardcoded backend

    if id is None:
        id = project_id

    from_id_fetch_remote_metadata(id,
                                  project_id=project_id,
                                  n_jobs=n_jobs,)


if __name__ == '__main__':
    from sparcur.simple.utils import pipe_main
    pipe_main(main)  # nothing to print or do after
# fetch_remote_metadata_all.py ends here



#!/usr/bin/env python3
# [[file:../../docs/developer-guide.org::fetch_secondary.py][fetch_secondary.py]]
from sparcur.paths import Path
from sparcur.simple.fetch_files import from_path_fetch_files, filter_manifests


def from_blob_fetch_files(dataset_blob, path=None):
    # should the blob contain a reference to the path
    # it was derived from?
    filter_function = filter_manifests(dataset_blob)
    from_path_fetch_files(path, filter_function, n_jobs=n_jobs)


def main(path=Path.cwd(), n_jobs=12, **kwargs):
    #breakpoint()  # XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
    # if not dataset_blob: get_blob vs path blob pairs?
    # starting from a partial blob means that this probably
    # should not kick off from the file system, but we know
    # that we will want to be able to kick it off from the
    # file system ... maybe the intermediate blobs can encode
    # the prov of where the file system reference they were
    # derived from lives ?
    dataset_blob = get_blob(path.cache_id)  # FIXME TODO
    from_blob_fetch_files(dataset_blob, path)


if __name__ == '__main__':
    from sparcur.simple.utils import pipe_main
    pipe_main(main)
# fetch_secondary.py ends here



""" Create a file with only the published subset curation-export.ttl

Usage:
    pushd path/to/export/root; python -m sparcur.export.published; popd

"""

import rdflib
from sparcur.paths import Path
from pyontutils.core import OntResPath, OntGraph
from pyontutils.namespaces import rdf, sparc, TEMP


def curation_export_published(export_path, out_base=None):
    p = Path(export_path).expanduser().resolve()
    ce = OntResPath(p / 'curation-export.ttl')
    orps = [OntResPath(_) for _ in (p / 'datasets').children if _.suffix == '.ttl']
    graphs = [o.graph for o in orps]

    merged = _populate_published(ce, graphs)

    op = p if out_base is None else Path(out_base)
    merged.write(op / 'curation-export-published.ttl')


def _merge_graphs(graphs):
    merged = OntGraph()
    for g in graphs:
        merged.namespace_manager.populate_from(
            {k:v for k, v in dict(g.namespace_manager).items()
            if k not in ('contributor', 'sample', 'subject')})
        merged.populate_from_triples(g.data)  # g.data excludes the owl:Ontology section
        # TODO switch the rdf:type of metadata section on combination to preserve export related metadata
    return merged


def _populate_published(curation_export, graphs):

    # datasets = [list(g[:rdf.type:sparc.Dataset]) for g in graphs]
    published_graphs = [
        g for g, doi in [(g, list(g[ds:TEMP.hasDoi]))
                        for g in graphs for ds in g[:rdf.type:sparc.Dataset]]
        if doi]

    merged = _merge_graphs(published_graphs)
    _fix_for_pub(curation_export, merged)
    return merged


def _fix_for_pub(curation_export, merged):
    mg = curation_export.metadata().graph
    mg.namespace_manager.populate(merged)

    new_bi = rdflib.URIRef(mg.boundIdentifier
                           .replace('ontologies/', 'ontologies/published/'))
    new_vi = rdflib.URIRef(mg.versionIdentifier
                           .replace('ontologies/', 'ontologies/published/'))
    replace_pairs = (
        (rdflib.Literal("SPARC Consortium curation export published graph"),
         rdflib.Literal("SPARC Consortium curation export graph")),
        (new_bi, mg.boundIdentifier),
        (new_vi, mg.versionIdentifier))

    new_meta = mg.replaceIdentifiers(replace_pairs)
    new_meta.populate(merged)
    return replace_pairs


def main():
    export_path = Path.cwd()
    curation_export_published(export_path)


if __name__ == '__main__':
    main()



""" convert dandi terms yaml to ttl """

import yaml
import rdflib
import augpathlib as aug
from pyontutils.core import populateFromJsonLd, OntGraph
from pyontutils.namespaces import rdfs, rdf

# pushd ~/git/NOFORK/dandi-schema/context
# python -m http.server 0 --bind 127.0.0.1
# get the tcp port from the python server (used as ${PORT} below)
# export PORT=
# sed -i "s/\.\.\/context\/base\.json/http:\/\/localhost:${PORT}\/base.json/" *.yaml

dandi = rdflib.Namespace('http://schema.dandiarchive.org/')
schema = rdflib.Namespace('http://schema.org/')


def path_yaml(string):
    with open(string, 'rb') as f:
        return yaml.safe_load(f)


def main():
    dandi_terms_path = aug.LocalPath.cwd()
    g = OntGraph()

    _ = [populateFromJsonLd(g, path_yaml(p))
         for p in dandi_terms_path.rglob('*.yaml')]
    g.write('dandi-raw.ttl')
    remove = [(s, p, o)
              for p in (schema.domainIncludes, schema.rangeIncludes, rdfs.subClassOf, rdf.type)
              for s, o in g[:p:]]
    add = [(s, p, (g.namespace_manager.expand(o.toPython()) if isinstance(o, rdflib.Literal) else o))
           for s, p, o in remove]
    _ = [g.remove(t) for t in remove]
    _ = [g.add(t) for t in add]
    # TODO ontology metadata header section
    g.write('dandi.ttl')


if __name__ == '__main__':
    main()



from sparcur.utils import PennsieveId
from sparcur.sparcron import get_redis_conn, state_lut


def dataset_status(conn, rawid):
    pid = PennsieveId(('dataset:' + rawid.split(':')[-1]))
    prefixes = 'state', 'updated', 'failed', 'sheet', 'verpi'
    keys = [f'{prefix}-{pid.id}' for prefix in prefixes]
    values = conn.mget(keys)
    out = {p:v for p, v in zip(prefixes, values)}
    out['id'] = pid.id
    out['state'] = state_lut[int(out['state'])]
    f = out['failed']
    out['failed'] = f.decode() if f else False
    out['sheet'] = int(out['sheet'])
    out['pipeline_internal_version'] = int(out.pop('verpi'))
    out['updated'] = out['updated'].decode()
    if out['failed'] and out['updated'] and out['failed'] < out['updated']:
        out['failed'] = False

    return out


def dataset_fails(conn):
    keys = conn.keys()
    _fkeys = [k for k in keys if b'failed-' in k]
    fvals = [v for v in conn.mget(_fkeys)]
    _fails = [(PennsieveId(('dataset:' + k.split(b':')[-1].decode())), v)
              for k, v in zip(_fkeys, fvals) if v]
    _ukeys = ['updated-N:dataset:' + i.uuid for i, _ in _fails]
    uvals = [v for v in conn.mget(_ukeys)]
    fails = [i for (i, f), u in zip(_fails, uvals) if not u or f > u]
    return fails


def main():
    from pprint import pprint
    conn = get_redis_conn()
    fails = dataset_fails(conn)
    if fails:
        pprint(fails)
        pprint(dataset_status(conn, fails[0].uuid))

    if False:
        pprint(dataset_status(
            conn, keys[0].split(b':')[-1].decode()))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# [[file:../../docs/developer-guide.org::disco.py][disco.py]]
import json
from pyontutils.core import OntGraph
from sparcur import export as ex
from sparcur.utils import PennsieveId
from sparcur.utils import fromJson, register_all_types  # FIXME this should also go in sparcron


def from_curation_export_json_path_xml_and_disco(curation_export_json_path):
    with open(curation_export_json_path, 'rt') as f:
        summary = json.load(f)

    datasets_json = summary['datasets']
    from_curation_export_json_path_datasets_json_xml_and_disco(
        curation_export_json_path, datasets_json)


def from_curation_export_json_path_datasets_json_xml_and_disco(
        curation_export_json_path, datasets_json, graphs=None):
    # FIXME need the snapshot linked somehow, export time started if we are using summary
    # or summary prov timestamp_export_start will give us the snapshot path as well if we
    # parse it back to a date
    if not graphs:
        snapshot_path = curation_export_json_path.parent / 'snapshot'
        paths = [(snapshot_path /
                  PennsieveId(d['id']).uuid /
                  'curation-export.ttl')
                 for d in datasets_json]
        graphs = [OntGraph().parse(path) for path in paths]

    datasets_ir = fromJson(datasets_json)
    ex.export_xml(curation_export_json_path, datasets_ir)
    ex.export_disco(curation_export_json_path, datasets_ir, graphs)
    # XXX not doing jsonld here, it will be combined
    # from single dataset jsonld or similar


def main(path=None, **kwargs):
    register_all_types()
    # FIXME the problem with this approach is that can cannot run
    # multiple downstream steps from the same upstream step, we would
    # need a compositional way to tell each downstream which upstreams
    # we wanted to run in any given situation, all to save additional
    # reads from disk
    if path is None:  # assume the user wants to run combine first
        from sparcur.simple.combine import main as combine
        curation_export_json_path, summary_all, _, graphs_all, _ = combine(**kwargs)
        datasets_json = summary_all['datasets']
        from_curation_export_json_path_datasets_json_xml_and_disco(
            curation_export_json_path, datasets_json, graphs)
    else:
        curation_export_json_path = path
        from_curation_export_json_path_xml_and_disco(curation_export_json_path)


if __name__ == '__main__':
    #import pprint
    from sparcur.simple.utils import pipe_main
    # these are really big, don't print them
    # pipe_main(main, after=pprint.pprint)
    pipe_main(main)
# disco.py ends here



from docopt import parse_defaults
from sparcur import exceptions as exc
from sparcur.cli import Report, Options, __doc__ as clidoc
from sparcur.paths import Path, BlackfynnCache
from sparcur.config import auth
from sparcur.server import make_app
from sparcur.backends import BlackfynnRemote
from sparcur.curation import Summary

defaults = {o.name:o.value if o.argcount else None
            for o in parse_defaults(clidoc)}

args = {'server': True,
        '--raw': False,
        '--latest': True,
        '--preview': False,  # set via
        '--sort-count-desc': True,
        '--tab-table': False,
        '<path>': [],
        '--verbose': False,
        '--to-sheets': False,

        #'--export-path': auth.get_path('export-path'),
        '--export-path': None,  # don't leak this
        '--partial': False,
        '--open': False,
        '--debug': False,

        '--export-file': None,  # when/where to pass this?
        '--published': True,  # lo and behold the solution! (hits export-url)
        '--ttl-file': None,  # FIXME TODO needed for terms
        '--ttl-compare': None,  # FIXME TODO needed for terms
        'hubmap': False,
        'hubmap-anatomy': False,

        '<project-id>': auth.get('remote-organization'),  # '--project-id':
        '--protcur-file': None,
        '--uri': True,  # needed by protcur export
        '--uri-html': True,  # use html link because share links are broken
        '--hypothesis-cache-file': None,
}

options = Options(args, defaults)
report = Report(options)

# set report paths that would normally be populated from Main
#report.cwd = options.project_path
#report.project_path = options.project_path
#report.project_id = project_path.cache.id  # FIXME should not have to do this manually?
#report.anchor = project_path.cache
#report.summary = Summary(options.project_path)
report._timestamp = None  # FIXME
report._folder_timestamp = None  # FIXME

# set up bfapi
#report.BlackfynnRemote = BlackfynnRemote._new(Path, BlackfynnCache)
#report.BlackfynnRemote.init(report.project_id)

app, *_ = make_app(report)
app.debug = False

if __name__ == '__main__':
    app.run(host='localhost', port=defaults['--port'], threaded=True)



#!/usr/bin/env python3
# [[file:../../docs/developer-guide.org::compact.py][compact.py]]
"""
Examples:
    # look at the second one and then run the first one since it is easier and safer to stop and resume
    # python -m sparcur.simple.compact | xargs -P4 -r -I{} sh -c 'tar -cvJf "{}.tar.xz" "{}" && rm -r "{}"'
    python -m sparcur.simple.compact | xargs -P12 -r -I{} echo tar -cvJf '{}.tar.xz' '{}'
    # python -m sparcur.simple.compact | xargs -P6 -r -I{} echo rm -r '{}'
"""
from sparcur.paths import Path
from sparcur.config import auth


__dep_cache = {}
def latest_snapped(dataset_export_path, snapped):
    if dataset_export_path not in __dep_cache:
        cs = set(c for c in dataset_export_path.children if c.is_dir() and not c.is_symlink())
        csnap = cs.intersection(snapped)
        if not csnap:  # no snap, pretend that latest is snapped
            # this can happen if there is no LATEST because we
            # were e.g. just exporting path metadata and nothing else
            maxsnap = sorted(cs, key=lambda p: p.name)[-1]
        else:
            maxsnap = sorted(csnap, key=lambda p: p.name)[-1]

        __dep_cache[dataset_export_path] = maxsnap.name

    return __dep_cache[dataset_export_path]


def main():
    export_path = Path(auth.get_path('export-path'))
    summary_path = export_path / 'summary'
    snapshots_path = export_path / 'snapshots'
    datasets_path = export_path / 'datasets'

    snap_shotted = [
        dss.resolve()
        for d in summary_path.rchildren_dirs
        for l in d.rchildren if l.is_symlink() and l.name == 'snapshot'
        for dss in l.resolve().children]

    snapped = set(snap_shotted)

    latest_sums = [
        d.resolve()
        for c in summary_path.children
        for d in (c / 'LATEST',) if d.exists()]

    all_builds = [
        build
        for date in datasets_path.children if date.is_dir() and not date.is_symlink()
        for build in date.children if build.is_dir() and not build.is_symlink()]

    older_not_snap = [
        a for a in all_builds
        if a not in snapped and a.name < latest_snapped(a.parent, snapped)]

    assert not snapped.intersection(older_not_snap)

    # newer = set(all_builds) - snapped - set(older_not_snap)

    _ = [print(p) for p in older_not_snap]


if __name__ == '__main__':
    main()
# compact.py ends here
